"""Tests for utils.dataframe."""

import numpy as np
import pandas as pd

from bycycle.utils.dataframes import *

###################################################################################################
###################################################################################################

def test_limit_df(sim_args):

    df_samples = sim_args['df_samples']
    fs = sim_args['fs']

    xlim = (1, 2)

    df_short = limit_df(df_samples, fs, start=xlim[0], stop=xlim[1])

    assert df_short['sample_next_trough'].min() >= 0
    assert df_short['sample_last_trough'].max() <= fs * (xlim[1] - xlim[0])


def test_get_extrema_df(sim_args):

    df_samples = sim_args['df_samples']
    center_e, side_e = get_extrema_df(df_samples)

    # The fixture will return peak centered cycles
    assert center_e == 'peak'
    assert side_e == 'trough'

    df_samples = pd.DataFrame({'sample_trough': []})
    center_e, side_e = get_extrema_df(df_samples)

    assert center_e == 'trough'
    assert side_e == 'peak'



"""Plot burst detection parameters."""

from itertools import cycle

import numpy as np
import pandas as pd
from scipy.stats import zscore
import matplotlib.pyplot as plt

from neurodsp.plts import plot_time_series, plot_bursts
from neurodsp.plts.utils import savefig

from bycycle.plts.cyclepoints import plot_cyclepoints_df
from bycycle.utils import limit_df, limit_signal, get_extrema_df
from bycycle.utils.checks import check_param

###################################################################################################
###################################################################################################

@savefig
def plot_burst_detect_summary(df_features, df_samples, sig, fs, threshold_kwargs, xlim=None,
                              figsize=(15, 3), plot_only_result=False, interp=True):
    """Plot the cycle-by-cycle burst detection parameters and burst detection summary.

    Parameters
    ----------
    df_features : pandas.DataFrame
        Dataframe output of :func:`~.compute_features`.
    df_samples : pandas.DataFrame
        Dataframe output of :func:`~.compute_cyclepoints`.
    sig : 1d array
        Time series to plot.
    fs : float
        Sampling rate, in Hz.
    threshold_kwargs : dict
        Burst parameter keys and threshold value pairs, as defined in the 'threshold_kwargs'
        argument of :func:`.compute_features`.
    xlim : tuple of (float, float), optional, default: None
        Start and stop times for plot.
    figsize : tuple of (float, float), optional, default: (15, 3)
        Size of each plot.
    plot_only_result : bool, optional, default: False
        Plot only the signal and bursts, excluding burst parameter plots.
    interp : bool, optional, default: True
        If True, interpolates between given values. Otherwise, plots in a step-wise fashion.

    Notes
    -----

    - If plot_only_result = True: return a plot of the burst detection in which periods with bursts
      are denoted in red.

    - If plot_only_result = False: return a list of the fig handle followed by the 5 axes.

    - In the top plot, the raw signal is plotted in black, and the red line indicates periods
      defined as oscillatory bursts. The highlighted regions indicate when each burst requirement
      was violated, color-coded consistently with the plots below.

      - blue: amp_fraction_threshold
      - red: amp_consistency_threshold
      - yellow: period_consistency_threshold
      - green: monotonicity_threshold

    Examples
    --------
    Plot the burst detection summary of a bursting signal:

    >>> from bycycle.features import compute_features
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> threshold_kwargs = {'amp_fraction_threshold': 0., 'amp_consistency_threshold': .5,
    ...                     'period_consistency_threshold': .5, 'monotonicity_threshold': .8}
    >>> df_features, df_samples = compute_features(sig, fs, f_range=(8, 12),
    ...                                            threshold_kwargs=threshold_kwargs)
    >>> plot_burst_detect_summary(df_features, df_samples, sig, fs, threshold_kwargs)
    """

    # Ensure arguments are within valid range
    check_param(fs, 'fs', (0, np.inf))

    # Normalize signal
    sig = zscore(sig)

    # Determine time array and limits
    times = np.arange(0, len(sig) / fs, 1 / fs)
    xlim = (times[0], times[-1]) if xlim is None else xlim

    # Determine if peak of troughs are the sides of an oscillation
    _, side_e = get_extrema_df(df_samples)

    # Remove this kwarg since it isn't stored cycle by cycle in the df (nothing to plot)
    if 'min_n_cycles' in threshold_kwargs.keys():
        del threshold_kwargs['min_n_cycles']

    n_kwargs = len(threshold_kwargs.keys())

    # Create figure and subplots
    if plot_only_result:
        fig, axes = plt.subplots(figsize=figsize, nrows=1)
        axes = [axes]
    else:
        fig, axes = plt.subplots(figsize=(figsize[0], figsize[1]*(n_kwargs+1)),
                                 nrows=n_kwargs+1, sharex=True)

    # Determine which samples are defined as bursting
    is_osc = np.zeros(len(sig), dtype=bool)
    df_osc = df_samples[df_features['is_burst']]

    for _, cyc in df_osc.iterrows():

        samp_start_burst = int(cyc['sample_last_' + side_e])
        samp_end_burst = int(cyc['sample_next_' + side_e] + 1)
        is_osc[samp_start_burst:samp_end_burst] = True

    # Plot bursts with extrema points
    xlabel = 'Time (s)' if len(axes) == 1 else ''

    plot_bursts(times, sig, is_osc, ax=axes[0], xlim=xlim, lw=2,
                labels=['Signal', 'Bursts'], xlabel='', ylabel='')

    plot_cyclepoints_df(df_samples, sig, fs, ax=axes[0], xlim=xlim, plot_zerox=False,
                        plot_sig=False, xlabel=xlabel, ylabel='Voltage\n(normalized)',
                        colors=['m', 'c'])

    # Plot each burst param
    colors = cycle(['blue', 'red', 'yellow', 'green', 'cyan', 'magenta', 'orange'])

    for idx, osc_key in enumerate(threshold_kwargs.keys()):

        column = osc_key.replace('_threshold', '')

        color = next(colors)

        # Highlight where a burst param falls below threshold
        for row_idx, cyc in df_samples.iterrows():

            if (df_features.iloc[row_idx][column] < threshold_kwargs[osc_key] and
                    df_features.iloc[row_idx]['is_burst'] == False):
                axes[0].axvspan(times[int(cyc['sample_last_' + side_e])],
                                times[int(cyc['sample_next_' + side_e])],
                                alpha=0.5, color=color, lw=0)

        # Plot each burst param on separate axes
        if not plot_only_result:

            ylabel = column.replace('_', ' ').capitalize()
            xlabel = 'Time (s)' if idx == n_kwargs-1 else ''

            plot_burst_detect_param(df_features, df_samples, sig, fs, column,
                                    threshold_kwargs[osc_key], figsize=figsize,
                                    ax=axes[idx+1], xlim=xlim, xlabel=xlabel, ylabel=ylabel,
                                    color=color, interp=interp)


@savefig
def plot_burst_detect_param(df_features, df_samples, sig, fs, burst_param, thresh,
                            xlim=None, ax=None, interp=True, **kwargs):
    """Plot a burst detection parameter and threshold.

    Parameters
    ----------
    df_features : pandas.DataFrame
        Dataframe output of :func:`~.compute_features`.
    df_samples : pandas.DataFrame
        Dataframe output of :func:`~.compute_shapes`.
    sig : 1d array
        Time series to plot.
    fs : float
        Sampling rate, in Hz.
    burst_param : str
        Column name of the parameter of interest in ``df``.
    thresh : float
        The burst parameter threshold. Parameter values greater
        than ``thresh`` are considered bursts.
    xlim : tuple of (float, float), optional, default: None
        Start and stop times for plot.
    ax : matplotlib.Axes, optional
        Figure axes upon which to plot.
    interp : bool, optional, default: True
        Interpolates points if true.
    **kwargs
        Keyword arguments to pass into `plot_time_series`.

    Notes
    -----
    Default keyword arguments include:

    - ``figsize``: tuple of (float, float), default: (15, 3)
    - ``xlabel``: str, default: 'Time (s)'
    - ``ylabel``: str, default: 'Voltage (uV)
    - ``color``: str, default: 'r'.

      - Note: ``color`` here is the fill color, rather than line color.

    Examples
    --------
    Plot the monotonicity of a bursting signal:

    >>> from bycycle.features import compute_features
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> threshold_kwargs = {'amp_fraction_threshold': 0., 'amp_consistency_threshold': .5,
    ...                     'period_consistency_threshold': .5, 'monotonicity_threshold': .8}
    >>> df_features, df_samples = compute_features(sig, fs, f_range=(8, 12),
    ...                                            threshold_kwargs=threshold_kwargs)
    >>> plot_burst_detect_param(df_features, df_samples, sig, fs, 'monotonicity', .8)
    """

    # Ensure arguments are within valid range
    check_param(fs, 'fs', (0, np.inf))

    # Set default kwargs
    figsize = kwargs.pop('figsize', (15, 3))
    xlabel = kwargs.pop('xlabel', 'Time (s)')
    ylabel = kwargs.pop('ylabel', burst_param)
    color = kwargs.pop('color', 'r')

    # Determine time array and limits
    times = np.arange(0, len(sig) / fs, 1 / fs)
    xlim = (times[0], times[-1]) if xlim is None else xlim

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    # Determine extrema strings
    center_e, side_e = get_extrema_df(df_samples)

    # Limit dataframe, sig and times
    df = pd.concat([df_samples, df_features], axis=1)

    df = limit_df(df, fs, start=xlim[0], stop=xlim[1])

    sig, times = limit_signal(times, sig, start=xlim[0], stop=xlim[1])

    # Remove start / end cycles that tlims falls between
    df = df[(df['sample_last_' + side_e] >= 0) & \
            (df['sample_next_' + side_e] < xlim[1]*fs)]

    # Plot burst param
    if interp:

        plot_time_series([times[df['sample_' + center_e]], xlim],
                         [df[burst_param], [thresh]*2], ax=ax, colors=['k', 'k'],
                         ls=['-', '--'], marker=["o", None], xlabel=xlabel,
                         ylabel="{0:s}\nthreshold={1:.2f}".format(ylabel, thresh), **kwargs)

    else:

        # Create steps, from side to side of each cycle, and set the y-value
        #   to the burst parameter value for that cycle
        side_times = np.array([])
        side_param = np.array([])

        for _, cyc in df.iterrows():

            # Get the times for the last and next side of a cycle
            side_times = np.append(side_times, [times[int(cyc['sample_last_' + side_e])],
                                                times[int(cyc['sample_next_' + side_e])]])

            # Set the y-value, from side to side, to the burst param for each cycle
            side_param = np.append(side_param, [cyc[burst_param]] * 2)

        plot_time_series([side_times, xlim], [side_param, [thresh]*2], ax=ax, colors=['k', 'k'],
                         ls=['-', '--'], marker=["o", None], xlim=xlim, xlabel=xlabel,
                         ylabel="{0:s}\nthreshold={1:.2f}".format(ylabel, thresh), **kwargs)

    # Highlight where param falls below threshold
    for _, cyc in df.iterrows():

        if cyc[burst_param] <= thresh:

            ax.axvspan(times[int(cyc['sample_last_' + side_e])],
                       times[int(cyc['sample_next_' + side_e])],
                       alpha=0.5, color=color, lw=0)



"""Plotting."""

from .burst import plot_burst_detect_summary, plot_burst_detect_param
from .cyclepoints import plot_cyclepoints_df, plot_cyclepoints_array
from .features import plot_feature_hist, plot_feature_categorical



"""Functions to compute features across 2 dimensional arrays of data."""

from functools import partial
from multiprocessing import Pool, cpu_count

import numpy as np

from bycycle.features import compute_features
from bycycle.group.utils import progress_bar

###################################################################################################
###################################################################################################

def compute_features_2d(sigs, fs, f_range, compute_features_kwargs=None,
                        return_samples=True, n_jobs=-1, progress=None):
    """Compute shape and burst features for a 2 dimensional array of signals.

    Parameters
    ----------
    sigs : 2d array
        Voltage time series, with shape [n_signals, n_points].
    fs : float
        Sampling rate, in Hz.
    f_range : tuple of (float, float)
        Frequency range for narrowband signal of interest, in Hz.
    compute_features_kwargs : dict or list of dict
        Keyword arguments used in :func:`~.compute_features`.
    return_samples : bool, optional, default: True
        Whether to return a dataframe of cyclepoint sample indices.
    n_jobs : int, optional, default: -1
        The number of jobs, one per cpu, to compute features in parallel.
    progress: {None, 'tqdm', 'tqdm.notebook'}, optional, default: None
        Specify whether to display a progress bar. Use 'tqdm' if installed.

    Returns
    -------
    df_features : list of pandas.DataFrame
        Dataframes containing shape and burst features for each cycle.
        Each dataframe is computed using the :func:`~.compute_features` function.
    df_samples : list of pandas.DataFrame, optional
        Dataframes containing cyclepoints for each cycle.
        Only returned if ``return_samples`` is True.
        Each dataframe is computed using the :func:`~.compute_features` function.

    Notes
    -----

    - The order of ``df_features`` and ``df_samples`` corresponds to the order of ``sigs``.
    - If ``compute_features_kwargs`` is a dictionary, the same kwargs are applied applied across
      the first axis of ``sigs``. Otherwise, a list of dictionaries equal in length to the
      first axis of ``sigs`` is required to apply unique kwargs to each signal.
    - ``return_samples`` is controlled from the kwargs passed in this function. If
      ``return_samples`` is a key in ``compute_features_kwargs``, it's value will be ignored.

    Examples
    --------
    Compute the features of a 2d array in parrallel using the same compute_features kwargs:

    >>> import numpy as np
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sigs = np.array([sim_bursty_oscillation(10, fs, 10) for i in range(10)])
    >>> compute_kwargs = {'burst_method': 'amp', 'threshold_kwargs':{'burst_fraction_threshold': 1}}
    >>> df_features = compute_features_2d(sigs, fs, f_range=(8, 12), return_samples=False, n_jobs=2,
    ...                                   compute_features_kwargs=compute_kwargs)

    Compute the features of a 2d array in parallel using using individualized settings per signal to
    examine the effect of various amplitude consistency thresholds:

    >>> sigs =  np.array([sim_bursty_oscillation(10, fs, freq=10)] * 10)
    >>> compute_kwargs = [{'threshold_kwargs': {'amp_consistency_threshold': thresh*.1}}
    ...                   for thresh in range(1, 11)]
    >>> df_features = compute_features_2d(sigs, fs, f_range=(8, 12), return_samples=False,
    ...                                   n_jobs=2, compute_features_kwargs=compute_kwargs)
    """

    if isinstance(compute_features_kwargs, list) and len(compute_features_kwargs) != len(sigs):
        raise ValueError("When compute_features_kwargs is a list, it's length must be equal to "
                         "sigs. Use a dictionary when applying the same kwargs to each signal.")
    elif compute_features_kwargs is None:
        compute_features_kwargs = {}

    # Drop `return_samples` argument, as this is set directly in the function call
    compute_features_kwargs = [compute_features_kwargs] if \
        isinstance(compute_features_kwargs, dict) else compute_features_kwargs

    [kwargs.pop('return_samples', None) for kwargs in compute_features_kwargs]

    n_jobs = cpu_count() if n_jobs == -1 else n_jobs

    with Pool(processes=n_jobs) as pool:

        if len(compute_features_kwargs) > 1:
            # Map iterable sigs and kwargs together
            mapping = pool.imap(partial(_proxy, fs=fs, f_range=f_range,
                                        return_samples=return_samples),
                                zip(sigs, compute_features_kwargs))

        else:
            # Only map sigs, kwargs are the same for each mapping
            mapping = pool.imap(partial(compute_features, fs=fs, f_range=f_range,
                                        return_samples=return_samples,
                                        **compute_features_kwargs[0]),
                                sigs)

        if return_samples is True:
            df_features, df_samples = zip(*progress_bar(mapping, progress, len(sigs)))

            df_features = list(df_features)
            df_samples = list(df_samples)

        else:
            df_features = list(progress_bar(mapping, progress, len(sigs)))

    if return_samples is True:
        return df_features, df_samples

    return df_features


def compute_features_3d(sigs, fs, f_range, compute_features_kwargs=None,
                        return_samples=True, n_jobs=-1, progress=None):
    """Compute shape and burst features for a 3 dimensional array of signals.

    Parameters
    ----------
    sigs : 3d array
        Voltage time series, with shape [n_groups, n_signals, n_points].
    fs : float
        Sampling rate, in Hz.
    f_range : tuple of (float, float)
        Frequency range for narrowband signal of interest, in Hz.
    compute_features_kwargs : dict or 1d list of dict or 2d list of dict
        Keyword arguments used in :func:`~.compute_features`.
    return_samples : bool, optional, default: True
        Whether to return a dataframe of cyclepoint sample indices.
    n_jobs : int, optional, default: -1
        The number of jobs, one per cpu, to compute features in parallel.
    progress: {None, 'tqdm', 'tqdm.notebook'}, optional, default: None
        Specify whether to display a progress bar. Use 'tqdm' if installed.

    Returns
    -------
    df_features : list of pandas.DataFrame
        Dataframes containing shape and burst features for each cycle.
        Each dataframe is computed using the :func:`~.compute_features` function.
    df_samples : list of pandas.DataFrame, optional
        Dataframes containing cyclepoints for each cycle.
        Only returned if ``return_samples`` is True.
        Each dataframe is computed using the :func:`~.compute_features` function.

    Notes
    -----

    - The order of ``df_features`` and ``df_samples`` corresponds to the order of ``sigs``.
    - If ``compute_features_kwargs`` is a dictionary, the same kwargs are applied applied across
      all signals. A 1d list, equal in length to the first dimensions of sigs, may be applied to
      each set of signals along the first dimensions. A 2d list, the same shape as the first two
      dimensions of ``sigs`` may also be used to applied unique parameters to each signal.
    - ``return_samples`` is controlled from the kwargs passed in this function. The
      ``return_samples`` value in ``compute_features_kwargs`` will be ignored.

    Examples
    --------
    Compute the features of a 3d array, in parallel, with a shape of
    (n_channels=2, n_epochs=3, n_signals=5000) using the same compute_features kwargs:

    >>> import numpy as np
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sigs = np.array([[sim_bursty_oscillation(10, fs, freq=10) for epoch in range(3)]
    ...                 for ch in range(2)])
    >>> threshold_kwargs = {'amp_consistency_threshold': .5, 'period_consistency_threshold': .5,
    ...                     'monotonicity_threshold': .8, 'min_n_cycles': 3}
    >>> compute_feature_kwargs = {'threshold_kwargs': threshold_kwargs, 'center_extrema': 'trough'}
    >>> features = compute_features_3d(sigs, fs, f_range= (8, 12), return_samples=False, n_jobs=2,
    ...                                compute_features_kwargs=compute_feature_kwargs)

    Compute the features of a 3d array, in parallel, with a shape of
    (n_channels=2, n_epochs=3, n_signals=5000) using channel-specific compute_features kwargs:

    >>> threshold_kwargs_ch1 = {'amp_consistency_threshold': .25, 'monotonicity_threshold': .25,
    ...                         'period_consistency_threshold': .25, 'min_n_cycles': 3}
    >>> threshold_kwargs_ch2 = {'amp_consistency_threshold': .5, 'monotonicity_threshold': .5,
    ...                         'period_consistency_threshold': .5, 'min_n_cycles': 3}
    >>> compute_kwargs = [{'threshold_kwargs': threshold_kwargs_ch1, 'center_extrema': 'trough'},
    ...                   {'threshold_kwargs': threshold_kwargs_ch2, 'center_extrema': 'trough'}]
    >>> features = compute_features_3d(sigs, fs, f_range= (8, 12), return_samples=False, n_jobs=2,
    ...                                compute_features_kwargs=compute_kwargs)
    """

    # Convert list of kwargs to array to check dimensions
    kwargs = compute_features_kwargs

    if isinstance(kwargs, list):
        kwargs = np.array(kwargs)
    elif kwargs is None:
        kwargs = {}

    # Ensure compute_features corresponds to sigs
    if isinstance(kwargs, np.ndarray):

        if kwargs.ndim == 1 and np.shape(kwargs)[0] != np.shape(sigs)[0]:

            raise ValueError("When compute_features_kwargs is a 1d list, it's length must "
                             "be equal to the first dimension of sigs. Use a dictionary "
                             "when applying the same kwargs to each signal.")

        elif ((kwargs.ndim == 2 and np.shape(kwargs)[0] != np.shape(sigs)[0]) or
              (kwargs.ndim == 2 and np.shape(kwargs)[1] != np.shape(sigs)[1])):

            raise ValueError("When compute_features_kwargs is a 2d list, it's shape must "
                             "be equal to the shape of the first two dimensions of sigs. "
                             "Use a dictionary when applying the same kwargs to each signal.")

    n_jobs = cpu_count() if n_jobs == -1 else n_jobs

    # Reshape sigs into 2d aray
    sigs_2d = sigs.reshape(np.shape(sigs)[0]*np.shape(sigs)[1], np.shape(sigs)[2])

    # Reshape kwargs
    if isinstance(kwargs, np.ndarray) and kwargs.ndim == 1:
        # Repeat each kwargs for each signal along the second dimension
        kwargs = np.array([[kwarg]*np.shape(sigs)[1] for kwarg in kwargs])

    # Flatten kwargs
    if isinstance(kwargs, np.ndarray):
        kwargs = list(kwargs.flatten())

    df_features = \
        compute_features_2d(sigs_2d, fs, f_range, compute_features_kwargs=kwargs,
                            return_samples=return_samples, n_jobs=n_jobs, progress=progress)

    if return_samples:

        df_features, df_samples = df_features[0], df_features[1]
        df_samples = _reshape_df(df_samples, sigs)

    # Reshape returned features to match the first two dimensions of sigs
    df_features = _reshape_df(df_features, sigs)

    if return_samples:
        return df_features, df_samples

    return df_features


def _proxy(args, fs=None, f_range=None, return_samples=None):
    """Proxy function to map kwargs and sigs together."""

    sig, kwargs = args[0], args[1:]
    return compute_features(sig, fs=fs, f_range=f_range,
                            return_samples=return_samples, **kwargs[0])


def _reshape_df(df_features, sigs_3d):
    """Reshape a list of dataframes."""

    df_reshape = []

    dim_b = np.shape(sigs_3d)[1]

    max_idx = [i for i in range(dim_b, len(df_features)*dim_b, dim_b)]
    min_idx = [i for i in range(0, len(df_features), dim_b)]

    for lower, upper in zip(min_idx, max_idx):
        df_reshape.append(df_features[lower:upper])

    return df_reshape



"""Functions to compute features across epoched data."""

from importlib import import_module

###################################################################################################
###################################################################################################

def progress_bar(iterable, progress, n_to_run):
    """Add a progress bar to an iterable to be processed.

    Parameters
    ----------
    iterable : list or iterable
        Iterable object to potentially apply progress tracking to.
    progress : {None, 'tqdm', 'tqdm.notebook'}, optional
        Which kind of progress bar to use. If None, no progress bar is used.
    n_to_run : int
        Number of jobs to complete.

    Returns
    -------
    pbar : iterable or tqdm object
        Iterable object, with tqdm progress functionality, if requested.

    Raises
    ------
    ValueError
        If the input for `progress` is not understood.

    Notes
    -----

    - ``tqdm`` must be installed separately from bycycle.
    - The explicit `n_to_run` input is required as tqdm requires this in the parallel case.
      The `tqdm` object that is potentially returned acts the same as the underlying iterable,
      with the addition of printing out progress every time items are requested.


    Examples
    --------
    Use a ``tqdm`` progress bar, which must me installed separately from ``bycycle``,
    when computing the features for 10 signals:

    >>> from multiprocessing import Pool
    >>> from functools import partial
    >>> from bycycle.features import compute_features
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> sigs = [sim_bursty_oscillation(10, fs=500, freq=10)] * 10
    >>> mapping = Pool(1).imap(partial(compute_features, fs=500, f_range=(8, 12),
    ...                                return_samples=False), sigs)
    >>> df_features = list(progress_bar(mapping, progress='tqdm', n_to_run=len(sigs)))
    """

    # Check progress specifier is okay
    tqdm_options = ['tqdm', 'tqdm.notebook']
    if progress is not None and progress not in tqdm_options:
        raise ValueError("Progress bar option not understood.")

    # Set the display text for the progress bar
    pbar_desc = 'Computing Bycycle Features'

    # Use a tqdm, progress bar, if requested
    if progress:

        # Try loading the tqdm module
        try:
            tqdm = import_module(progress)

            # If tqdm loaded, apply the progress bar to the iterable
            pbar = tqdm.tqdm(iterable, desc=pbar_desc, total=n_to_run, dynamic_ncols=True)

        except ImportError:

            # If tqdm isn't available, proceed without a progress bar
            print(("A progress bar requiring the 'tqdm' module was requested, "
                   "but 'tqdm' is not installed. \nProceeding without using a progress bar."))
            pbar = iterable

    # If progress is None, return the original iterable without a progress bar applied
    else:
        pbar = iterable

    return pbar



"""Detect bursts: amplitude threshold approach."""

from bycycle.utils.checks import check_param
from bycycle.burst.utils import check_min_burst_cycles

###################################################################################################
###################################################################################################

def detect_bursts_amp(df_features, burst_fraction_threshold=1, min_n_cycles=3):
    """Detect bursts based on an amplitude thresholding.

    Parameters
    ----------
    df_features : pandas.DataFrame
        Waveform features for individual cycles from :func:`~.compute_burst_features`.
    burst_fraction_threshold : int or float, optional, default: 1
        Minimum fraction of a cycle to be identified as a burst.
    min_n_cycles : int, optional, default: 3
        The minimum number of cycles of consecutive cycles required to be considered a burst.

    Returns
    -------
    df_features : pandas.DataFrame
        Dataframe updated, with a additional column to indicate if the cycle is part of a burst.

    Examples
    --------
    Apply thresholding for dual amplitude burst detection:

    >>> from bycycle.features import compute_burst_features, compute_shape_features
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> df_shapes, df_samples = compute_shape_features(sig, fs, f_range=(8, 12))
    >>> df_burst = compute_burst_features(df_shapes, df_samples, sig, burst_method='amp',
    ...                                   burst_kwargs={'fs': fs, 'f_range': (8, 12)})
    >>> df_burst = detect_bursts_amp(df_burst)
    """

    # Ensure arguments are within valid ranges
    check_param(burst_fraction_threshold, 'burst_fraction_threshold', (0, 1))

    # Determine cycles that are defined as bursting throughout the whole cycle
    is_burst = [frac >= burst_fraction_threshold for frac in df_features['burst_fraction']]

    df_features['is_burst'] = check_min_burst_cycles(is_burst, min_n_cycles=min_n_cycles)

    return df_features



"""Tests plts.features."""

import numpy as np

from bycycle.burst import detect_bursts_cycles

from bycycle.tests.tutils import plot_test
from bycycle.tests.settings import TEST_PLOTS_PATH

from bycycle.plts.features import *

###################################################################################################
###################################################################################################

@plot_test
def test_plot_feature_hist(sim_args):

    df_features = sim_args['df_features']
    threshold_kwargs = sim_args['threshold_kwargs']

    df_features = detect_bursts_cycles(df_features, **threshold_kwargs)

    plot_feature_hist(df_features, 'amp_consistency', xlim=(0, 1), save_fig=True,
                      file_name='test_plot_feature_hist', file_path=TEST_PLOTS_PATH)

@plot_test
def test_plot_feature_categorical(sim_args):

    df_features = sim_args['df_features']
    threshold_kwargs = sim_args['threshold_kwargs']

    df_features = detect_bursts_cycles(df_features, **threshold_kwargs)

    # Compare first & second halves of the signal - distributions should be the same
    group = np.array(['First Half' for row in range(len(df_features))])
    group[:round(len(group)/2)] = 'Second Half'
    df_features['group'] = group

    plot_feature_categorical(df_features, 'amp_consistency', group_by='group', save_fig=True,
                             file_name='test_plot_feature_hist', file_path=TEST_PLOTS_PATH)



"""Tests for burst.cycle."""

import itertools

import numpy as np

from bycycle.features import compute_burst_features

from bycycle.burst.cycle import *

###################################################################################################
###################################################################################################

def test_detect_bursts_cycles(sim_args):

    df_features = sim_args['df_features']
    threshold_kwargs = sim_args['threshold_kwargs']

    # Apply consistency burst detection
    df_burst_cycles = detect_bursts_cycles(df_features, **threshold_kwargs)

    # Make sure that burst detection is only boolean
    assert df_burst_cycles.dtypes['is_burst'] == 'bool'
    assert df_burst_cycles['is_burst'].mean() >= 0
    assert df_burst_cycles['is_burst'].mean() <= 1
    assert np.min([sum(1 for _ in group) for key, group in \
        itertools.groupby(df_burst_cycles['is_burst']) if key]) >= 3



"""Plot extrema and zero-crossings."""

import numpy as np
import matplotlib.pyplot as plt

from neurodsp.plts import plot_time_series
from neurodsp.plts.utils import savefig

from bycycle.utils.checks import check_param
from bycycle.utils import limit_signal, get_extrema_df

###################################################################################################
###################################################################################################

@savefig
def plot_cyclepoints_df(df_samples, sig, fs, plot_sig=True, plot_extrema=True,
                        plot_zerox=True, xlim=None, ax=None, **kwargs):
    """Plot extrema and/or zero-crossings from a DataFrame.

    Parameters
    ----------
    df_samples: pandas.DataFrame
        Dataframe output of :func:`~.compute_cyclepoints`.
    sig : 1d array
        Time series to plot.
    fs : float
        Sampling rate, in Hz.
    plot_sig : bool, optional, default: True
        Whether to also plot the raw signal.
    plot_extrema :  bool, optional, default: True
        Whether to plots the peaks and troughs.
    plot_zerox :  bool, optional, default: True
        Whether to plots the zero-crossings.
    xlim : tuple of (float, float), optional
        Start and stop times.
    ax : matplotlib.Axes, optional
        Figure axes upon which to plot.
    **kwargs
        Keyword arguments to pass into `plot_time_series`.

    Notes
    -----
    Default keyword arguments include:

    - ``figsize``: tuple of (float, float), default: (15, 3)
    - ``xlabel``: str, default: 'Time (s)'
    - ``ylabel``: str, default: 'Voltage (uV)

    Examples
    --------
    Plot cyclepoints using a dataframe from :func:`~.compute_cyclepoints`:

    >>> from bycycle.features import compute_cyclepoints
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> df_samples = compute_cyclepoints(sig, fs, f_range=(8, 12))
    >>> plot_cyclepoints_df(df_samples, sig, fs)
    """

    # Ensure arguments are within valid range
    check_param(fs, 'fs', (0, np.inf))

    # Determine extrema/zero-crossings from dataframe
    center_e, side_e = get_extrema_df(df_samples)

    peaks, troughs, rises, decays = [None]*4

    if plot_extrema:

        peaks = df_samples['sample_' + center_e].values
        troughs = np.append(df_samples['sample_last_' + side_e].values,
                            df_samples['sample_next_' + side_e].values[-1])
    if plot_zerox:

        rises = df_samples['sample_zerox_rise'].values
        decays = df_samples['sample_zerox_decay'].values

    plot_cyclepoints_array(sig, fs, peaks=peaks, troughs=troughs, rises=rises,
                           decays=decays, plot_sig=plot_sig, xlim=xlim, ax=ax, **kwargs)


@savefig
def plot_cyclepoints_array(sig, fs, peaks=None, troughs=None, rises=None, decays=None,
                           plot_sig=True, xlim=None, ax=None, **kwargs):
    """Plot extrema and/or zero-crossings from arrays.

    Parameters
    ----------
    sig : 1d array
        Time series to plot.
    fs : float
        Sampling rate, in Hz.
    peaks : 1d array, optional
        Peak signal indices from :func:`.find_extrema`.
    troughs : 1d array, optional
        Trough signal indices from :func:`.find_extrema`.
    rises : 1d array, optional
        Zero-crossing rise indices from :func:`~.find_zerox`.
    decays : 1d array, optional
        Zero-crossing decay indices from :func:`~.find_zerox`.
    plot_sig : bool, optional, default: True
        Whether to also plot the raw signal.
    xlim : tuple of (float, float), optional
        Start and stop times.
    ax : matplotlib.Axes, optional, default: None
        Figure axes upon which to plot.
    **kwargs
        Keyword arguments to pass into `plot_time_series`.

    Notes
    -----
    Default keyword arguments include:

    - ``figsize``: tuple of (float, float), default: (15, 3)
    - ``xlabel``: str, default: 'Time (s)'
    - ``ylabel``: str, default: 'Voltage (uV)
    - ``colors``: list, default: ['k', 'b', 'r', 'g', 'm']

    Examples
    --------
    Plot cyclepoints using arrays from :func:`.find_extrema` and  :func:`~.find_zerox`:

    >>> from bycycle.cyclepoints import find_extrema, find_zerox
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> peaks, troughs = find_extrema(sig, fs, f_range=(8, 12), boundary=0)
    >>> rises, decays = find_zerox(sig, peaks, troughs)
    >>> plot_cyclepoints_array(sig, fs, peaks=peaks, troughs=troughs, rises=rises, decays=decays)
    """

    # Ensure arguments are within valid range
    check_param(fs, 'fs', (0, np.inf))

    # Set times and limits
    times = np.arange(0, len(sig) / fs, 1 / fs)
    xlim = (times[0], times[-1]) if xlim is None else xlim

    # Restrict sig and times to xlim
    sig, times = limit_signal(times, sig, start=xlim[0], stop=xlim[1])

    # Set default kwargs
    figsize = kwargs.pop('figsize', (15, 3))
    xlabel = kwargs.pop('xlabel', 'Time (s)')
    ylabel = kwargs.pop('ylabel', 'Voltage (uV)')
    default_colors = ['b', 'r', 'g', 'm']

    # Extend plotting based on given arguments
    x_values = []
    y_values = []
    colors = ['k']

    for idx, points in enumerate([peaks, troughs, rises, decays]):

        if points is not None:

            # Limit times and shift indices of cyclepoints (cps)
            cps = points[(points >= xlim[0]*fs) & (points < xlim[1]*fs)]
            cps = cps - int(xlim[0]*fs)

            y_values.append(sig[cps])
            x_values.append(times[cps])
            colors.append(default_colors[idx])

    # Allow custom colors to overwrite default
    colors = kwargs.pop('colors', colors)

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    if plot_sig:
        plot_time_series(times, sig, colors=colors[0], ax=ax)
        colors = colors[1:]

    plot_time_series(x_values, y_values, ax=ax, xlabel=xlabel, ylabel=ylabel,
                     colors=colors, marker='o', ls='', **kwargs)



"""Utilities for burst detection."""

from copy import deepcopy
import numpy as np
from bycycle.utils.checks import check_param

###################################################################################################
###################################################################################################

def check_min_burst_cycles(is_burst, min_n_cycles=3):
    """Enforce minimum number of consecutive cycles to be considered a burst.

    Parameters
    ----------
    is_burst : 1d array
        Boolean array indicating which cycles are bursting.
    min_n_cycles : int, optional, default: 3
        The minimum number of cycles of consecutive cycles required to be considered a burst.

    Returns
    -------
    is_burst : 1d array
        Updated burst array.

    Examples
    --------
    Remove bursts with less than 3 consectutive cycles:

    >>> is_burst = np.array([False, True, True, False, True, True, True, True, False])
    >>> check_min_burst_cycles(is_burst)
    array([False, False, False, False,  True,  True,  True,  True, False])
    """

    # Ensure argument is within valid range
    check_param(min_n_cycles, 'min_n_cycles', (0, np.inf))

    temp_cycle_count = 0

    for idx, bursting in enumerate(is_burst):

        if bursting:
            temp_cycle_count += 1

        else:

            if temp_cycle_count < min_n_cycles:
                for c_rm in range(temp_cycle_count):
                    is_burst[idx - 1 - c_rm] = False

            temp_cycle_count = 0

    return is_burst


def recompute_edges(df_features, threshold_kwargs):
    """Recompute the is_burst column for cycles on the edges of bursts.

    Parameters
    ----------
    df_features : pandas.DataFrame
        A dataframe containing shape and burst features for each cycle.
    threshold_kwargs : dict, optional, default: None
        Feature thresholds for cycles to be considered bursts, matching keyword arguments for:

        - :func:`~.detect_bursts_cycles` for consistency burst detection
          (i.e. when burst_method == 'cycles')

    Returns
    -------
    df_features_edges : pandas.DataFrame
        An cycle feature dataframe with an updated ``is_burst`` column for edge cycles.

    Notes
    -----

    - `df_features` must be computed using consistency burst detection.

    Examples
    --------
    Lower the amplitude consistency threshold to zero for cycles on the edges of bursts:

    >>> sig = sim_combined(n_seconds=4, fs=1000, components={'sim_bursty_oscillation': {'freq': 10},
    ...                                                      'sim_powerlaw': {'exp': 2}})
    >>> threshold_kwargs = {'amp_fraction_threshold': 0., 'amp_consistency_threshold': .5,
    ...                     'period_consistency_threshold': .5, 'monotonicity_threshold': .4,
    ...                     'min_n_cycles': 3}
    >>> df_features, _ = compute_features(sig, fs=1000, f_range=(8, 12), threshold_kwargs=threshold_kwargs)
    >>> threshold_kwargs['amp_consistency_threshold'] = 0
    >>> df_features_edges = recompute_edges(df_features, threshold_kwargs)
    """

    # Prevent circular import between burst.utils and burst.cycle
    from bycycle.burst import detect_bursts_cycles

    # Prevent overwriting the orignal dataframe
    df_features_edges = df_features.copy()

    # Identify all cycles where is_burst changes on the following cycle
    #   Use copy to keep dataframe columns unlinked
    is_burst = deepcopy(df_features_edges['is_burst'].values)
    burst_edges = np.where(is_burst[1:] == ~is_burst[:-1])[0]

    # Adjust odd edges such that all edges fall on is_burst == False
    burst_edges = np.array([edge if idx % 2 == 0 else edge+1 for idx, edge in
                            enumerate(burst_edges)])

    # Recompute is_burst
    df_features_edges =  detect_bursts_cycles(
        df_features_edges,
        amp_fraction_threshold=threshold_kwargs['amp_fraction_threshold'],
        amp_consistency_threshold=threshold_kwargs['amp_consistency_threshold'],
        period_consistency_threshold=threshold_kwargs['period_consistency_threshold'],
        monotonicity_threshold=threshold_kwargs['monotonicity_threshold'],
        min_n_cycles=threshold_kwargs['min_n_cycles']
    )

    # Confine recomputed is_burst to edges
    is_burst[burst_edges] = df_features_edges.iloc[burst_edges]['is_burst'].values

    df_features_edges['is_burst'] = is_burst

    return df_features_edges



"""Tests for cyclepoints.zerox."""

import os

import numpy as np

from bycycle.cyclepoints import find_extrema

from bycycle.cyclepoints.zerox import *

###################################################################################################
###################################################################################################

def test_find_zerox(sim_stationary):
    """Test ability to find peaks and troughs."""

    # Load signal
    sig = sim_stationary

    fs = 1000
    f_range = (6, 14)

    # Find peaks and troughs
    peaks, troughs = find_extrema(sig, fs, f_range, boundary=1, first_extrema='peak')

    # Find zerocrossings
    rises, decays = find_zerox(sig, peaks, troughs)

    assert len(peaks) == (len(rises) + 1)
    assert len(troughs) == len(decays)
    assert peaks[0] < decays[0]
    assert decays[0] < troughs[0]
    assert troughs[0] < rises[0]
    assert rises[0] < peaks[1]



"""Utility functions for working with ByCycle DataFrames."""

import numpy as np

from bycycle.utils.checks import check_param

###################################################################################################
###################################################################################################

def limit_df(df, fs, start=None, stop=None):
    """Restrict dataframe to be within time limits.

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe output of :func:`~.compute_features`.
    fs : float
        Sampling rate, in Hz.
    start : float, optional
        The lower time limit, in seconds, to restrict the df.
    stop : float, optional
        The upper time limit, in seconds, to restrict the df.

    Returns
    -------
    df : pandas.DataFrame
        A limited dataframe of cycle features.

    Notes
    -----
    Cycles, or rows in the `df`, are included if any segment of the cycle falls after the
    `stop` time or before the `end` time.

    Examples
    --------
    Limit a samples dataframe to the first second of a simulated signal:

    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> from bycycle.features import compute_features
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> _, df_samples = compute_features(sig, fs, f_range=(8, 12))
    >>> df_samples_limited = limit_df(df_samples, fs, start=0, stop=1)
    """

    # Ensure arguments are within valid range
    check_param(fs, 'fs', (0, np.inf))
    check_param(start, 'start', (0, stop))
    check_param(stop, 'stop', (start, np.inf))

    center_e, side_e = get_extrema_df(df)

    start = 0 if start is None else start

    df = df[df['sample_next_' + side_e].values >= start*fs]

    if stop is not None:
        df = df[df['sample_last_' + side_e].values < stop*fs]

    # Shift sample indices to start at 0
    df['sample_last_' + side_e] = df['sample_last_' + side_e] - int(fs * start)
    df['sample_next_' + side_e] = df['sample_next_' + side_e] - int(fs * start)
    df['sample_' + center_e] = df['sample_' + center_e] - int(fs * start)
    df['sample_zerox_rise'] = df['sample_zerox_rise'] - int(fs * start)
    df['sample_zerox_decay'] = df['sample_zerox_decay'] - int(fs * start)

    return df


def get_extrema_df(df):
    """Determine whether cycles are peak or trough centered.

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe output of :func:`~.compute_features`.

    Returns
    -------
    center_e : str
        Center extrema, either 'peak' or 'trough'.
    side_e : str
        Side extrema, either 'peak' or 'trough'.

    Examples
    --------
    Confirm that cycles are peak-centered:

    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> from bycycle.features import compute_features
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> _, df_samples = compute_features(sig, fs, f_range=(8, 12), center_extrema='peak')
    >>> center_e, side_e = get_extrema_df(df_samples)
    >>> center_e
    'peak'
    """

    center_e = 'peak' if 'sample_peak' in df.columns else 'trough'
    side_e = 'trough' if center_e == 'peak' else 'peak'

    return center_e, side_e


def rename_extrema_df(center_extrema, df_samples, df_features):
    """Rename a dataframe based on the centered extrema.

    Parameters
    ----------
    center_extrema : {'trough', 'peak'}
        Which extrema is centered.
    df_samples, df_features : pandas.DataFrames
        ByCycle dataframes to rename, given the centered extrema.

    Returns
    -------
    df_features, df_samples : pandas.DataFrames
        Updated dataframes.

    Examples
    --------
    Convert the column labels of a peak-centered dataframe to a trough-centered dataframe:

    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> from bycycle.features import compute_features
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> sig = -sig  # invert the signal, flipping peaks and troughs
    >>> df_features, df_samples = compute_features(sig, fs, f_range=(8, 12), center_extrema='peak')
    >>> df_features, df_samples = rename_extrema_df('trough', df_samples, df_features)
    """

    # Rename columns if they are actually trough-centered
    if center_extrema == 'trough':

        samples_rename_dict = {'sample_peak': 'sample_trough',
                               'sample_zerox_decay': 'sample_zerox_rise',
                               'sample_zerox_rise': 'sample_zerox_decay',
                               'sample_last_trough': 'sample_last_peak',
                               'sample_next_trough': 'sample_next_peak'}

        features_rename_dict = {'time_peak': 'time_trough',
                                'time_trough': 'time_peak',
                                'volt_peak': 'volt_trough',
                                'volt_trough': 'volt_peak',
                                'time_rise': 'time_decay',
                                'time_decay': 'time_rise',
                                'volt_rise': 'volt_decay',
                                'volt_decay': 'volt_rise'}

        df_samples.rename(columns=samples_rename_dict, inplace=True)
        df_features.rename(columns=features_rename_dict, inplace=True)

        # Need to reverse symmetry measures
        df_features['volt_peak'] = -df_features['volt_peak']
        df_features['volt_trough'] = -df_features['volt_trough']
        df_features['time_rdsym'] = 1 - df_features['time_rdsym']
        df_features['time_ptsym'] = 1 - df_features['time_ptsym']

    return df_features, df_samples



"""Utility functions for working with time series."""

import numpy as np

from bycycle.utils.checks import check_param

###################################################################################################
###################################################################################################

def limit_signal(times, sig, start=None, stop=None):
    """Restrict signal and times to be within time limits.

    Parameters
    ----------
    times : 1d array
        Time definition for the time series.
    sig : 1d array
        Time series.
    start : float
        The lower time limit, in seconds, to restrict the signal.
    stop : float
        The upper time limit, in seconds, to restrict the signal.

    Returns
    -------
    sig : 1d array
        A limited time series.
    times : 1d array
        A limited time definition.

    Examples
    --------
    Restrict a signal and times to the first second:

    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> from neurodsp.utils import create_times
    >>> sig = sim_bursty_oscillation(n_seconds=10, fs=500, freq=10)
    >>> times = create_times(n_seconds=10, fs=500)
    >>> sig, times = limit_signal(times, sig, start=0, stop=1)

    """

    # Ensure arguments are within valid range
    check_param(start, 'start', (0, stop))
    check_param(stop, 'stop', (start, np.inf))

    if start is not None:
        sig = sig[times >= start]
        times = times[times >= start]

    if stop is not None:
        sig = sig[times < stop]
        times = times[times < stop]

    return sig, times



"""Test burst.utils."""

import numpy as np

from bycycle.features import compute_features
from bycycle.burst.utils import *


###################################################################################################
###################################################################################################

def test_check_min_burst_cycles():

    is_burst = np.array([True, True, True, False])
    is_burst_check = check_min_burst_cycles(is_burst, min_n_cycles=3)

    assert (is_burst == is_burst_check).all()

    is_burst = np.array([True, False, True, False])
    is_burst_check = check_min_burst_cycles(is_burst, min_n_cycles=3)

    assert not any(is_burst_check)


def test_recompute_edges(sim_args_comb):

    # Grab sim arguments from fixture
    sig = sim_args_comb['sig']
    threshold_kwargs = sim_args_comb['threshold_kwargs']
    fs = sim_args_comb['fs']
    n_seconds = len(sig) / fs
    f_range = sim_args_comb['f_range']
    df_features = sim_args_comb['df_features']

    # Case 1: use the same thresholds should result in the same dataframe
    df_features_edges = recompute_edges(df_features, threshold_kwargs)
    assert (df_features['is_burst'].values == df_features['is_burst'].values).all()

    # Case 2: ensure the number of burst increases when edge thresholds are lowered

    # This guarantees at least one non-burst cycle on the edge will be recomputed
    #   as a burst by duplicating the signal and adding zero-padding between
    sig_force_recomp = np.zeros(2*len(sig) + fs)
    sig_force_recomp[:int(fs *  n_seconds)] = sig
    sig_force_recomp[int(fs * n_seconds) + fs:] = sig

    df_features = compute_features(sig, fs, f_range, threshold_kwargs=threshold_kwargs,
                                   return_samples=False)

    # Update thresholds to give all edge cycles is_burst = True, except for the first and
    #   last cycles, since these cycles contain np.nan values for consistency measures
    threshold_kwargs['amp_fraction_threshold'] = 0
    threshold_kwargs['amp_consistency_threshold'] = 0
    threshold_kwargs['period_consistency_threshold'] = 0
    threshold_kwargs['monotonicity_threshold'] = 0

    df_features_edges = recompute_edges(df_features, threshold_kwargs)

    # Ensure that at least one cycle was added to is_burst after recomputed
    is_burst_orig = len(np.where(df_features['is_burst'] == True)[0])
    is_burst_recompute = len(np.where(df_features_edges['is_burst'] == True)[0])

    assert is_burst_recompute > is_burst_orig



"""Interpolated phase, using cyclepoints."""

import numpy as np

###################################################################################################
###################################################################################################

def extrema_interpolated_phase(sig, peaks, troughs, rises=None, decays=None):
    """Use extrema and (optionally) zero-crossings to estimate instantaneous phase.

    Parameters
    ----------
    sig : 1d array
        Time series.
    peaks : 1d array
        Samples of oscillatory peaks.
    troughs : 1d array
        Samples of oscillatory troughs.
    rises : 1d array, optional
        Samples of oscillatory rising zero-crossings.
    decays : 1d array, optional
        Samples of oscillatory decaying zero-crossings.

    Returns
    -------
    pha : 1d array
        Instantaneous phase time series.

    Notes
    -----
    - Phase is encoded as:
        - phase 0 for peaks
        - phase pi/2 for decay zero-crossing
        - phase pi/-pi for troughs
        - phase -pi/2 for rise zero-crossing
    - Extrema and zero-crossing estimation can be poor if, for example, the signal is noisy.
      In such cases, the same index may be assigned to both a peak and a decaying zero-crossing.
      To address this, we first assign phase values by zero-crossings, and then may overwrite
      them with extrema phases.
    - Using burst detection helps avoid analyzing oscillatory properties of
      non-oscillatory sections of the signal.

    Examples
    --------
    Estimate phase from peaks and troughs:

    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> from bycycle.cyclepoints import find_extrema
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> peaks, troughs = find_extrema(sig, fs, f_range=(8, 12))
    >>> pha = extrema_interpolated_phase(sig, peaks, troughs)
    """

    # Initialize phase arrays, one for trough pi and trough -pi
    sig_len = len(sig)
    times = np.arange(sig_len)
    pha_tpi = np.zeros(sig_len) * np.nan
    pha_tnpi = np.zeros(sig_len) * np.nan

    # If specified, assign phases to zero-crossings
    if rises is not None:
        pha_tpi[rises] = -np.pi / 2
        pha_tnpi[rises] = -np.pi / 2
    if decays is not None:
        pha_tpi[decays] = np.pi / 2
        pha_tnpi[decays] = np.pi / 2

    # Define phases
    pha_tpi[peaks] = 0
    pha_tpi[troughs] = np.pi
    pha_tnpi[peaks] = 0
    pha_tnpi[troughs] = -np.pi

    # Interpolate to find all phases
    pha_tpi = np.interp(times, times[~np.isnan(pha_tpi)], pha_tpi[~np.isnan(pha_tpi)])
    pha_tnpi = np.interp(times, times[~np.isnan(pha_tnpi)], pha_tnpi[~np.isnan(pha_tnpi)])

    pha = _merge_phases(pha_tpi, pha_tnpi)

    return pha


def _merge_phases(pha_tpi, pha_tnpi):
    """Helper functions for extrema_interpolated_phase."""

    # Create phase differences to determine where phase is decaying/rising in the -pi array
    diffs = np.diff(pha_tnpi)

    # Pad the phase difference array with a NaN to maintain a length equal to the timeseries
    diffs = np.append(diffs, np.nan)

    # Create new phase series, using trough pi for decaying periods & trough -pi for rising periods
    pha = np.array([pha_tpi[idx] if diffs[idx] < 0 else pha for idx, pha in enumerate(pha_tnpi)])

    # Assign the periods before the first empirical phase timepoint to NaN
    diffs = np.diff(pha)
    first_empirical_idx = next(idx for idx, xi in enumerate(diffs) if xi > 0)
    pha[:first_empirical_idx] = np.nan

    # Assign the periods after the last empirical phase timepoint to NaN
    diffs = np.diff(pha)
    last_empirical_idx = next(idx for idx, xi in enumerate(diffs[::-1]) if xi > 0)
    pha[-last_empirical_idx + 1:] = np.nan

    return pha



"""Plot cycle features."""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from neurodsp.plts.utils import savefig

###################################################################################################
###################################################################################################

@savefig
def plot_feature_hist(feature, param_label, only_bursts=True, bins='auto', ax=None, **kwargs):
    """Plot a histogram for a cycle feature.

    Parameters
    ----------
    feature : pandas.DataFrame or 1d array
        Dataframe output from :func:`~.compute_features` or a 1d array.
    param_label : str
        Column name of the parameter of interest in ``df_features``.
    only_burst : bool, optional, default: True
        Whether to limit cycles to only those that are bursting.
    bins : int or string, optional, default: 'auto'
        The number of bins or binning strategy, as specified in `matplotlib.pyplot.hist`.
    ax : matplotlib.Axes, optional
        Figure axes upon which to plot.
    **kwargs
        Keyword arguments to pass into `matplotlib.pyplot.hist`.

    Notes
    -----
    Default keyword arguments include:

    - ``xlabel``: str, default: ``param_label``
    - ``figsize``: tuple of (float, float), default: (10, 10)
    - ``color``: str, default: 'k'
    - ``xlim``: tuple of (float, float), default: None
    - ``fontsize``: float, default: 15
    - ``alpha``: float, default: .5

    Examples
    --------
    Plot a histogram of each cycle's mean band amplitude:

    >>> from bycycle.features import compute_features
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> df_features = compute_features(sig, fs, f_range=(8, 12), return_samples=False)
    >>> plot_feature_hist(df_features, 'band_amp', only_bursts=False)
    """

    # Limit dataframe to bursts
    if isinstance(feature, pd.core.frame.DataFrame) and only_bursts is True:
        feature = feature[feature['is_burst']][param_label]
    elif isinstance(feature, pd.core.frame.DataFrame) and only_bursts is False:
        feature = feature[param_label]

    # Default keyword args
    figsize = kwargs.pop('figsize', (5, 5))
    xlabel = kwargs.pop('xlabel', param_label)
    xlim = kwargs.pop('xlim', None)
    fontsize = kwargs.pop('fontsize', 15)
    alpha = kwargs.pop('alpha', .5)

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    # Axis style
    ax.set_xlabel(xlabel, size=fontsize)
    ax.set_ylabel('# cycles', size=fontsize)

    if xlim:
        ax.set_xlim(xlim)

    ax.hist(feature, bins=bins, alpha=alpha, **kwargs)

    if 'label' in kwargs:
        ax.legend(fontsize=fontsize)


@savefig
def plot_feature_categorical(df_features, param_label, group_by=None, ax=None, **kwargs):
    """Plot a cycle feature by one or more categories.

    Parameters
    ----------
    df_features : pandas.DataFrame
        Dataframe output from :func:`~.compute_features`.
    param_label : str
        Column name of the parameter of interest in ``df_features``.
    group_by : str, optional
        Dataframe column name of a grouping variable to split plotting by.
    ax : matplotlib.Axes, optional
        Figure axes upon which to plot.
    **kwargs
        Keyword arguments to pass into `matplotlib.pyplot.scatter`.

    Notes
    -----
    Default keyword arguments include:

    - ``xlabel``: list, default: [1, 2, 3...]
    - ``ylabel``: str, default: ``param_label``
    - ``figsize``: tuple of (float, float), default: (10, 10)
    - ``fontsize``: float, default: 20

    Examples
    --------
    Plot and compare the rise-decay times of two asine signals:

    >>> import numpy as np
    >>> import pandas as pd
    >>> from neurodsp.sim import sim_oscillation
    >>> from bycycle.group import compute_features_2d
    >>> fs = 500
    >>> sigs = np.array([sim_oscillation(5, fs, 10, cycle='asine', rdsym=0.2),
    ...                  sim_oscillation(5, fs, 10, cycle='asine', rdsym=0.8)])
    >>> features = compute_features_2d(sigs, fs, f_range=(8, 12), return_samples=False, n_jobs=2)
    >>> features[0]['group'], features[1]['group'] = 'low', 'high'
    >>> df_features = pd.concat([features[0], features[1]])
    >>> plot_feature_categorical(df_features, 'time_rdsym', group_by='group')
    """

    # Split features by group if specified
    features = [df_features[param_label]]

    if group_by is not None:

        features = []

        for group in np.unique(df_features[group_by].values):
            features.append(df_features[df_features[group_by] == group][param_label])

    # Add random variance along x-axis
    x_values = [np.random.normal(idx+1, 0.05, len(feature)) for idx, feature in enumerate(features)]

    # Default keyword args
    figsize = kwargs.pop('figsize', (10, 10))
    ylabel = kwargs.pop('ylabel', param_label)
    xlabel = kwargs.pop('xlabel', [idx+1 for idx in range(len(features))])
    fontsize = kwargs.pop('fontsize', 20)

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    for idx, feature in enumerate(features):
        ax.scatter(x_values[idx], feature, **kwargs)

    # Plot styling
    ax.set_xlim(.5, len(features)+.5)
    ax.set_xticks([idx+1 for idx in range(len(features))])
    ax.set_xticklabels(xlabel)
    ax.set_ylabel(ylabel, fontsize=fontsize)
    ax.tick_params(axis='both', which='major', labelsize=fontsize)

    # Hide the right and top spines
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)



"""Configuration file for pytest for bycycle."""

import os
import shutil

import pytest

from neurodsp.sim import sim_oscillation, sim_combined
from neurodsp.utils.sim import set_random_seed

from bycycle.features import compute_shape_features, compute_burst_features, compute_features
from bycycle.tests.settings import (N_SECONDS, FS, FREQ, F_RANGE,
                                    BASE_TEST_FILE_PATH, TEST_PLOTS_PATH)

###################################################################################################
###################################################################################################

def pytest_configure(config):

    set_random_seed(42)


@pytest.fixture(scope='module')
def sim_args():

    sig = sim_oscillation(N_SECONDS, FS, FREQ)

    df_shapes, df_samples = compute_shape_features(sig, FS, F_RANGE, return_samples=True)
    df_burst = compute_burst_features(df_shapes, df_samples, sig)
    df_features, df_samples = compute_features(sig, FS, F_RANGE, return_samples=True)

    threshold_kwargs = {'amp_fraction_threshold': 0.,
                        'amp_consistency_threshold': .5,
                        'period_consistency_threshold': .5,
                        'monotonicity_threshold': .5,
                        'min_n_cycles': 3}

    yield {'sig': sig, 'fs': FS, 'f_range': F_RANGE, 'df_features': df_features,
           'df_shapes': df_shapes, 'df_burst': df_burst, 'df_samples': df_samples,
           'threshold_kwargs': threshold_kwargs}


@pytest.fixture(scope='module')
def sim_args_comb():

    components = {'sim_bursty_oscillation': {'freq': FREQ}, 'sim_powerlaw': {'exp': 2}}

    sig = sim_combined(N_SECONDS, FS, components=components)

    df_shapes, df_samples = compute_shape_features(sig, FS, F_RANGE, return_samples=True)
    df_burst = compute_burst_features(df_shapes, df_samples, sig)
    df_features, df_samples = compute_features(sig, FS, F_RANGE, return_samples=True)

    threshold_kwargs = {'amp_fraction_threshold': 0.,
                        'amp_consistency_threshold': .5,
                        'period_consistency_threshold': .5,
                        'monotonicity_threshold': .5,
                        'min_n_cycles': 3}

    yield {'sig': sig, 'fs': FS, 'f_range': F_RANGE, 'df_features': df_features,
           'df_shapes': df_shapes, 'df_burst': df_burst, 'df_samples': df_samples,
           'threshold_kwargs': threshold_kwargs}


@pytest.fixture(scope='session', autouse=True)
def check_dir():
    """Once, prior to session, this will clear and re-initialize the test file directories."""

    # If the directories already exist, clear them
    if os.path.exists(BASE_TEST_FILE_PATH):
        shutil.rmtree(BASE_TEST_FILE_PATH)

    # Remake (empty) directories
    os.mkdir(BASE_TEST_FILE_PATH)
    os.mkdir(TEST_PLOTS_PATH)


@pytest.fixture(scope='module')
def sim_stationary():

    sig = sim_oscillation(N_SECONDS, FS, FREQ, phase=0.15,
                          cycle="asine", rdsym=.3)
    yield sig




"""Tests for features.burst."""

import pytest

import numpy as np

from bycycle.features import *

###################################################################################################
###################################################################################################

@pytest.mark.parametrize("dual_thresh", [True, False])
@pytest.mark.parametrize("center_e", ['peak', 'trough'])
def test_compute_burst_features(sim_args, dual_thresh, center_e):

    sig = sim_args['sig']
    df_shape_features = sim_args['df_shapes']
    df_samples = sim_args['df_samples']

    if center_e == 'trough':

        # Swap column names (fixture is peak centered)
        rename_dict = {'sample_peak': 'sample_trough',
                       'sample_zerox_decay': 'sample_zerox_rise',
                       'sample_zerox_rise': 'sample_zerox_decay',
                       'sample_last_trough': 'sample_last_peak',
                       'sample_next_trough': 'sample_next_peak'}

        df_samples.rename(columns=rename_dict, inplace=True)

    if dual_thresh:

        # Use dual threshold burst detecion.
        burst_detection_kwargs =  {'fs': sim_args['fs'], 'f_range': sim_args['f_range']}

        df_burst_features = compute_burst_features(df_shape_features, df_samples, sig,
                                                   burst_method='amp',
                                                   burst_kwargs=burst_detection_kwargs)

        burst_fraction = df_burst_features['burst_fraction']

        assert np.nan not in burst_fraction

        assert np.all((burst_fraction >= 0) & (burst_fraction <= 1))

    else:

        # Use consistency burst detection
        df_burst_features = compute_burst_features(df_shape_features, df_samples, sig)

        amp_fraction = df_burst_features['amp_fraction'].values[1:-1]
        amp_consistency = df_burst_features['amp_consistency'].values[1:-1]
        period_consistency = df_burst_features['period_consistency'].values[1:-1]
        monotonicity = df_burst_features['monotonicity'].values[1:-1]

        assert np.all((amp_fraction >= 0) & (amp_fraction <= 1))
        assert np.all((amp_consistency >= 0) & (amp_consistency <= 1))
        assert np.all((period_consistency >= 0) & (period_consistency <= 1))
        assert np.all((monotonicity >= 0) & (monotonicity <= 1))

    assert len(df_shape_features) == len(df_burst_features)



"""Tests for plts.burst."""

import pytest

import numpy as np

from bycycle.burst import detect_bursts_cycles

from bycycle.tests.tutils import plot_test
from bycycle.tests.settings import TEST_PLOTS_PATH

from bycycle.plts import *

###################################################################################################
###################################################################################################

@plot_test
@pytest.mark.parametrize("interp", [True, False])
def test_plot_burst_detect_param(sim_args, interp):

    df_samples = sim_args['df_samples']
    df_features = sim_args['df_features']
    sig = sim_args['sig']
    fs = sim_args['fs']

    thresh = np.nanmin(df_features['amp_consistency'].values) + 0.1

    plot_burst_detect_param(df_features, df_samples, sig, fs, 'amp_consistency',
                            thresh, interp=interp, save_fig=True,
                            file_path=TEST_PLOTS_PATH, file_name='test_plot_burst_detect_param')


@plot_test
@pytest.mark.parametrize("plot_only_result", [True, False])
def test_plot_burst_detect_summary(sim_args, plot_only_result):

    burst_detection_kwargs = {'amp_fraction_threshold': 1,
                              'amp_consistency_threshold': .5,
                              'period_consistency_threshold': .5,
                              'monotonicity_threshold': .8,
                              'min_n_cycles': 3}

    df_samples = sim_args['df_samples']
    df_features = sim_args['df_features']
    sig = sim_args['sig']
    fs = sim_args['fs']

    df_features = detect_bursts_cycles(df_features, **burst_detection_kwargs)

    plot_burst_detect_summary(df_features, df_samples, sig, fs, burst_detection_kwargs,
                              plot_only_result=plot_only_result,
                              save_fig=True, file_path=TEST_PLOTS_PATH,
                              file_name='test_plot_burst_detect_summary')



"""Tests for burst.amp."""

import itertools
import numpy as np

from bycycle.features import compute_burst_features

from bycycle.burst.amp import *

###################################################################################################
###################################################################################################

def test_detect_bursts_amp(sim_args):

    df_shape_features = sim_args['df_shapes']
    df_samples = sim_args['df_samples']
    sig = sim_args['sig']
    burst_kwarg = {'fs': sim_args['fs'], 'f_range': sim_args['f_range'], 'amp_threshes': (0.5, 1)}

    df_features = compute_burst_features(df_shape_features, df_samples, sig,
                                         burst_method='amp',
                                         burst_kwargs=burst_kwarg)

    # Apply dual threshold burst detection
    df_features = detect_bursts_amp(df_features, burst_fraction_threshold=1, min_n_cycles=3)

    # Make sure that burst detection is only boolean
    assert df_features.dtypes['is_burst'] == 'bool'
    assert df_features['is_burst'].mean() >= 0
    assert df_features['is_burst'].mean() <= 1
    assert np.min([sum(1 for _ in group) for key, group \
        in itertools.groupby(df_features['is_burst']) if key]) >= 3



"""Compute burst features for individual cycles."""

import numpy as np
import pandas as pd

from bycycle.utils.checks import check_param
from bycycle.burst import detect_bursts_dual_threshold

###################################################################################################
###################################################################################################

def compute_burst_features(df_shape_features, df_samples, sig,
                           burst_method='cycles', burst_kwargs=None):
    """Compute burst features for each cycle.

    Parameters
    ----------
    df_shape_features : pandas.DataFrame
        Shape parameters for each cycle, determined using :func:`~.compute_shape_features`.
    df_samples : pandas.DataFrame
        Indices of cyclepoints returned from :func:`~.compute_cyclepoints`.
    sig : 1d array
        Voltage time series used for determining monotonicity.
    burst_method : string, optional, default: 'cycles'
        Method for detecting bursts.

        - 'cycles': detect bursts based on the consistency of consecutive periods & amplitudes
        - 'amp': detect bursts using an amplitude threshold

    burst_kwargs : dict, optional, default: None
        Additional arguments required for amplitude burst detection. Defined in
        :func:`~.compute_burst_fraction`, keys include:

        - ``fs`` : required for dual amplitude threshold burst detection
        - ``f_range`` : required for dual amplitude threshold burst detection
        - ``amp_threshes`` : optional, default: (1, 2)
        - ``min_n_cycles`` : optional, default: 3
        - ``filter_kwargs`` : optional, default: None

    Returns
    -------
    df_burst_features : pandas.DataFrame
        Dataframe containing burst features. Each row is one cycle. Columns:

        When cycle consistency burst detection is used (i.e. burst_method == 'cycles'):

        - ``amp_fraction`` : normalized amplitude
        - ``amp_consistency`` : difference in the rise and decay voltage within a cycle
        - ``period_consistency`` : difference between a cycle’s period and the period of the
          adjacent cycles
        - ``monotonicity`` : fraction of instantaneous voltage changes between consecutive
          samples that are positive during the rise phase and negative during the decay phase

        When dual threshold burst detection is used (i.e. burst_method == 'amp'):

        - ``burst_fraction`` : fraction of a cycle that is bursting

    Examples
    --------
    Compute burst features:

    >>> from bycycle.features import compute_shape_features
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs  = 500
    >>> sig = sim_bursty_oscillation(10, fs, 10)
    >>> df_shapes, df_samples = compute_shape_features(sig, fs, f_range=(8, 12))
    >>> df_burst = compute_burst_features(df_shapes, df_samples, sig, burst_method='amp',
    ...                                   burst_kwargs={'fs': fs, 'f_range': (8, 12)})
    """

    df_burst_features = pd.DataFrame()

    # Use feature consistency burst detection
    if burst_method == 'cycles':

        # Custom feature functions may be inserted here as long as an array is returned
        #   with length equal to the number of cycles, or rows in df_shapes
        df_burst_features['amp_fraction'] = compute_amp_fraction(df_shape_features)

        df_burst_features['amp_consistency'] = \
                compute_amp_consistency(df_shape_features, df_samples)

        df_burst_features['period_consistency'] = compute_period_consistency(df_shape_features)
        df_burst_features['monotonicity'] = compute_monotonicity(df_samples, sig)

    # Use dual threshold burst detection
    elif burst_method == 'amp':

        fs = burst_kwargs.pop('fs', None)
        f_range = burst_kwargs.pop('f_range', None)

        if fs is None or f_range is None:
            raise ValueError("'fs' and 'f_range' must be defined in 'burst_kwargs' "
                             "when 'burst_method' is 'amp'.")

        df_burst_features['burst_fraction'] = \
            compute_burst_fraction(df_samples, sig, fs, f_range, **burst_kwargs)

    else:
        raise ValueError("Unrecognized 'burst_method'.")

    return df_burst_features


def compute_amp_fraction(df_shape_features):
    """Compute the amplitude fraction of each cycle.

    Parameters
    ----------
    df_shape_features : pandas.DataFrame
        Shape features for each cycle, determined using :func:`~.compute_shape_features`.

    Returns
    -------
    amp_fract : 1d array
        The amplitude fraction of each cycle.

    Examples
    --------
    Compute amplitude fractions.

    >>> from bycycle.features import compute_shape_features
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> df_shapes, df_samples = compute_shape_features(sig, fs, (8, 12))
    >>> amp_fraction = compute_amp_fraction(df_shapes)
    """

    return df_shape_features['volt_amp'].rank() / len(df_shape_features)


def compute_amp_consistency(df_shape_features, df_samples):
    """Compute amplitude consistency for each cycle.

    Parameters
    ----------
    df_shape_features : pandas.DataFrame
        Shape features for each cycle, determined using :func:`~.compute_shape_features`.
    df_samples : pandas.DataFrame
        Indices of cyclepoints returned from :func:`~.compute_cyclepoints`.

    Returns
    -------
    amp_consist : 1d array
        The amplitude consistency of each cycle.

    Examples
    --------
    Compute amplitude consistency:

    >>> from bycycle.features import compute_shape_features
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> df_shapes, df_samples = compute_shape_features(sig, fs, f_range=(8, 12))
    >>> amp_consistency = compute_amp_consistency(df_shapes, df_samples)
    """

    # Compute amplitude consistency
    cycles = len(df_shape_features)
    amp_consistency = np.ones(cycles) * np.nan
    rises = df_shape_features['volt_rise'].values
    decays = df_shape_features['volt_decay'].values

    for cyc in range(1, cycles-1):

        # Division by zero will return np.nan, supress warning.
        with np.errstate(invalid='ignore', divide='ignore'):

            consist_current = np.min([rises[cyc], decays[cyc]]) / np.max([rises[cyc], decays[cyc]])

            if 'sample_peak' in df_samples.columns:

                consist_last = np.min([rises[cyc], decays[cyc-1]]) / \
                    np.max([rises[cyc], decays[cyc-1]])

                consist_next = np.min([rises[cyc+1], decays[cyc]]) / \
                    np.max([rises[cyc+1], decays[cyc]])

            else:

                consist_last = np.min([rises[cyc-1], decays[cyc]]) / \
                    np.max([rises[cyc-1], decays[cyc]])

                consist_next = np.min([rises[cyc], decays[cyc+1]]) / \
                    np.max([rises[cyc], decays[cyc+1]])

            if np.isnan([consist_current, consist_next, consist_last]).all():
                amp_consistency[cyc] = np.nan
            else:
                amp_consistency[cyc] = np.nanmin([consist_current, consist_next, consist_last])

    return amp_consistency


def compute_period_consistency(df_shape_features):
    """Compute the period consistency of each cycle.

    Parameters
    ----------
    df_shape_features : pandas.DataFrame
        Shape features for each cycle, determined using :func:`~.compute_shape_features`.

    Returns
    -------
    period_consistency : 1d array
        The period consistency of each cycle.

    Examples
    --------
    Compute period consistency:

    >>> from bycycle.features import compute_shape_features
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> df_shapes, df_samples = compute_shape_features(sig, fs, f_range=(8, 12))
    >>> period_consistency = compute_period_consistency(df_shapes)
    """

    # Compute period consistency
    cycles = len(df_shape_features)
    period_consistency = np.ones(cycles) * np.nan
    periods = df_shape_features['period'].values

    for cyc in range(1, cycles-1):

        consist_last = np.min([periods[cyc], periods[cyc-1]]) / \
            np.max([periods[cyc], periods[cyc-1]])
        consist_next = np.min([periods[cyc+1], periods[cyc]]) / \
            np.max([periods[cyc+1], periods[cyc]])

        period_consistency[cyc] = np.min([consist_next, consist_last])

    return period_consistency


def compute_monotonicity(df_samples, sig):
    """Compute the monotonicity of each cycle.

    Parameters
    ----------
    df_samples : pandas.DataFrame
        Indices of cyclepoints returned from :func:`~.compute_cyclepoints`.
    sig : 1d array
        Time series.

    Returns
    -------
    monotonicity : 1d array
        The monotonicity of each cycle.

    Examples
    --------
    Compute monotonicity:

    >>> from bycycle.features import compute_cyclepoints
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> df_samples = compute_cyclepoints(sig, fs, f_range=(8, 12))
    >>> montonicity = compute_monotonicity(df_samples, sig)
    """

    # Compute monotonicity
    cycles = len(df_samples)
    monotonicity = np.ones(cycles) * np.nan

    for idx, row in df_samples.iterrows():

        if 'sample_peak' in df_samples.columns:
            rise_period = sig[int(row['sample_last_trough']):int(row['sample_peak'])+1]
            decay_period = sig[int(row['sample_peak']):int(row['sample_next_trough'])+1]

        else:
            decay_period = sig[int(row['sample_last_peak']):int(row['sample_trough'])+1]
            rise_period = sig[int(row['sample_trough']):int(row['sample_next_peak'])+1]

        decay_mono = np.mean(np.diff(decay_period) < 0)
        rise_mono = np.mean(np.diff(rise_period) > 0)
        monotonicity[idx] = np.mean([decay_mono, rise_mono])

    return monotonicity


def compute_burst_fraction(df_samples, sig, fs, f_range, amp_threshes=(1, 2),
                           min_n_cycles=3, filter_kwargs=None):
    """Compute the proportion of each cycle that is bursting using a dual threshold algorithm.

    Parameters
    ----------
    df_samples : pandas.DataFrame
        Indices of cyclepoints returned from :func:`~.compute_cyclepoints`.
    sig : 1d array
        Time series.
    fs : float
        Sampling rate, Hz.
    f_range : tuple of (float, float)
        Frequency range (Hz) for oscillator of interest.
    amp_threshes : tuple (low, high), optional, default: (1, 2)
        Threshold values for determining timing of bursts.
        These values are in units of amplitude (or power, if specified) normalized to
        the median amplitude (value 1).
    min_n_cycles : int, optional, default: 3
        Minimum number of cycles to be identified as truly oscillating needed in a row in
        order for them to remain identified as truly oscillating.
    filter_kwargs : dict, optional, default: None
        Keyword arguments to :func:`~neurodsp.filt.filter.filter_signal`.

    Returns
    -------
    burst_fraction : 1d array
        The proportion of each cycle that is bursting.

    Notes
    -----
    If a cycle contains three samples and the corresponding section of `is_burst` is
    np.array([True, True, False]), the burst fraction is 0.66 for that cycle.

    Examples
    --------
    Compute proportions of cycles that are bursting using dual amplitude thresholding:

    >>> from bycycle.features import compute_cyclepoints
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> df_samples = compute_cyclepoints(sig, fs, f_range=(8, 12))
    >>> burst_fraction = compute_burst_fraction(df_samples, sig, fs, f_range=(8, 12))
    """

    # Ensure arguments are within valid ranges
    check_param(fs, 'fs', (0, np.inf))
    check_param(amp_threshes[0], 'lower amp_threshes', (0, amp_threshes[1]))
    check_param(amp_threshes[1], 'upper amp_threshes', (amp_threshes[0], np.inf))

    filter_kwargs = {} if filter_kwargs is None else filter_kwargs

    # Detect bursts using the dual amplitude threshold approach
    is_burst = detect_bursts_dual_threshold(sig, fs, amp_threshes, f_range,
                                            min_n_cycles=min_n_cycles, **filter_kwargs)

    # Convert the boolean array to binary
    is_burst = is_burst.astype(int)

    # Determine cycle sides
    side_e = 'trough' if 'sample_peak' in df_samples.columns else 'peak'

    # Compute fraction of each cycle that's bursting
    burst_fraction = []
    for _, row in df_samples.iterrows():
        fraction_bursting = np.mean(is_burst[int(row['sample_last_' + side_e]):
                                             int(row['sample_next_' + side_e] + 1)])
        burst_fraction.append(fraction_bursting)

    return burst_fraction



"""Compute cycle-by-cycle features."""

import warnings
import numpy as np
import pandas as pd

from bycycle.utils.checks import check_param
from bycycle.features.shape import compute_shape_features
from bycycle.features.burst import compute_burst_features
from bycycle.burst import detect_bursts_cycles, detect_bursts_amp

###################################################################################################
###################################################################################################

def compute_features(sig, fs, f_range, center_extrema='peak', burst_method='cycles',
                     burst_kwargs=None, threshold_kwargs=None, find_extrema_kwargs=None,
                     return_samples=True):
    """Compute shape and burst features for each cycle.

    Parameters
    ----------
    sig : 1d array
        Time series.
    fs : float
        Sampling rate, in Hz.
    f_range : tuple of (float, float)
        Frequency range for narrowband signal of interest (Hz).
    center_extrema : {'peak', 'trough'}
        The center extrema in the cycle.

        - 'peak' : cycles are defined trough-to-trough
        - 'trough' : cycles are defined peak-to-peak

    burst_method : string, optional, default: 'cycles'
        Method for detecting bursts.

        - 'cycles': detect bursts based on the consistency of consecutive periods & amplitudes
        - 'amp': detect bursts using an amplitude threshold

    burst_kwargs : dict, optional, default: None
        Additional keyword arguments defined in :func:`~.compute_burst_fraction` for dual
        amplitude threshold burst detection (i.e. when burst_method == 'amp').
    threshold_kwargs : dict, optional, default: None
        Feature thresholds for cycles to be considered bursts, matching keyword arguments for:

        - :func:`~.detect_bursts_cycles` for consistency burst detection
          (i.e. when burst_method == 'cycles')
        - :func:`~.detect_bursts_amp` for  amplitude threshold burst detection
          (i.e. when burst_method == 'amp').

    find_extrema_kwargs : dict, optional, default: None
        Keyword arguments for function to find peaks an troughs (:func:`~.find_extrema`)
        to change filter Parameters or boundary. By default, it sets the filter length to three
        cycles of the low cutoff frequency (``f_range[0]``).
    return_samples : bool, optional, default: True
        Returns samples indices of cyclepoints used for determining features if True.

    Returns
    -------
    df_features : pandas.DataFrame
        A dataframe containing shape and burst features for each cycle. Columns:

        - ``period`` : period of the cycle
        - ``time_decay`` : time between peak and next trough
        - ``time_rise`` : time between peak and previous trough
        - ``time_peak`` : time between rise and decay zero-crosses
        - ``time_trough`` : duration of previous trough estimated by zero-crossings
        - ``volt_decay`` : voltage change between peak and next trough
        - ``volt_rise`` : voltage change between peak and previous trough
        - ``volt_amp`` : average of rise and decay voltage
        - ``volt_peak`` : voltage at the peak
        - ``volt_trough`` : voltage at the last trough
        - ``time_rdsym`` : fraction of cycle in the rise period
        - ``time_ptsym`` : fraction of cycle in the peak period
        - ``band_amp`` : average analytic amplitude of the oscillation.

        When consistency burst detection is used (i.e. burst_method == 'cycles'):

        - ``amp_fraction`` : normalized amplitude
        - ``amp_consistency`` : difference in the rise and decay voltage within a cycle
        - ``period_consistency`` : difference between a cycle’s period and the period of the
          adjacent cycles
        - ``monotonicity`` : fraction of instantaneous voltage changes between consecutive
          samples that are positive during the rise phase and negative during the decay phase

        When dual threshold burst detection is used (i.e. burst_method == 'amp'):

        - ``burst_fraction`` : fraction of a cycle that is bursting

    df_samples : pandas.DataFrame, optional, default: True
        An optionally returned dataframe containing cyclepoints for each cycle.
        Columns (listed for peak-centered cycles):

        - ``sample_peak`` : sample of 'sig' at which the peak occurs
        - ``sample_zerox_decay`` : sample of the decaying zero-crossing
        - ``sample_zerox_rise`` : sample of the rising zero-crossing
        - ``sample_last_trough`` : sample of the last trough
        - ``sample_next_trough`` : sample of the next trough

    Examples
    --------
    Compute shape and burst features:

    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> df_features, df_samples = compute_features(sig, fs, f_range=(8, 12))
    """

    # Ensure arguments are within valid range
    check_param(fs, 'fs', (0, np.inf))

    # Compute shape features for each cycle.
    df_shape_features, df_samples = \
        compute_shape_features(sig, fs, f_range, center_extrema=center_extrema,
                               find_extrema_kwargs=find_extrema_kwargs)

    # Ensure kwargs are a dictionaries
    if burst_method == 'amp' and not isinstance(burst_kwargs, dict):
        burst_kwargs = {}

    if not isinstance(threshold_kwargs, dict):
        threshold_kwargs = {}
        warnings.warn("""
            No burst detection thresholds are provided. This is not recommended. Please
            inspect your data and choose appropriate parameters for 'threshold_kwargs'.
            Default burst detection parameters are likely not well suited for your
            desired application.
            """)

    # Ensure required kwargs are set for amplitude burst detection
    if burst_method == 'amp':
        burst_kwargs['fs'] = fs
        burst_kwargs['f_range'] = f_range

    # Compute burst features for each cycle
    df_burst_features = compute_burst_features(df_shape_features, df_samples, sig,
                                               burst_method=burst_method,
                                               burst_kwargs=burst_kwargs)

    # Concatenate shape and burst features
    df_features = pd.concat((df_shape_features, df_burst_features), axis=1)

    # Define whether or not each cycle is part of a burst
    if burst_method == 'cycles':
        df_features = detect_bursts_cycles(df_features, **threshold_kwargs)
    elif burst_method == 'amp':
        df_features = detect_bursts_amp(df_features, **threshold_kwargs)
    else:
        raise ValueError('Invalid argument for "burst_method".'
                         'Either "cycles" or "amp" must be specified."')

    if return_samples:
        return df_features, df_samples

    return df_features



"""Compute cyclepoint features for individual cycles."""

import pandas as pd
import numpy as np

from bycycle.utils.checks import check_param
from bycycle.cyclepoints import find_extrema, find_zerox

###################################################################################################
###################################################################################################

def compute_cyclepoints(sig, fs, f_range, **find_extrema_kwargs):
    """Compute sample indices for cyclepoints.

    Parameters
    ----------
    sig : 1d array
        Time series.
    fs : float
        Sampling rate, in Hz.
    f_range : tuple of (float, float)
        Frequency range, in Hz, to narrowband filter the signal, used to find zero-crossings.
    find_extrema_kwargs : dict, optional, default: None
        Keyword arguments for the function to find peaks and troughs (:func:`~.find_extrema`)
        that change filter parameters or boundary. By default, the boundary is set to zero.

    Returns
    -------
    df_samples : pandas.DataFrame, optional, default: False
        Dataframe containing sample indices of cyclepoints.
        Columns (listed for peak-centered cycles):

        - ``peaks`` : signal indices of oscillatory peaks
        - ``troughs`` :  signal indices of oscillatory troughs
        - ``rises`` : signal indices of oscillatory rising zero-crossings
        - ``decays`` : signal indices of oscillatory decaying zero-crossings
        - ``sample_peak`` : sample of 'sig' at which the peak occurs
        - ``sample_zerox_decay`` : sample of the decaying zero-crossing
        - ``sample_zerox_rise`` : sample of the rising zero-crossing
        - ``sample_last_trough`` : sample of the last trough
        - ``sample_next_trough`` : sample of the next trough

    Examples
    --------
    Compute the signal indices of cyclepoints:

    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> df_samples = compute_cyclepoints(sig, fs, f_range=(8, 12))
    """

    # Ensure arguments are within valid range
    check_param(fs, 'fs', (0, np.inf))

    # Find extrema and zero-crossings locations in the signal
    peaks, troughs = find_extrema(sig, fs, f_range, **find_extrema_kwargs)
    rises, decays = find_zerox(sig, peaks, troughs)

    # For each cycle, identify the sample of each extrema and zero-crossing
    samples = {}
    samples['sample_peak'] = peaks[1:]
    samples['sample_last_zerox_decay'] = decays[:-1]
    samples['sample_zerox_decay'] = decays[1:]
    samples['sample_zerox_rise'] = rises
    samples['sample_last_trough'] = troughs[:-1]
    samples['sample_next_trough'] = troughs[1:]

    df_samples = pd.DataFrame.from_dict(samples)

    return df_samples



"""Tests for plts.cyclepoints."""

from bycycle.cyclepoints import find_extrema, find_zerox

from bycycle.tests.tutils import plot_test
from bycycle.tests.settings import TEST_PLOTS_PATH

from bycycle.plts.cyclepoints import *

###################################################################################################
###################################################################################################

@plot_test
def test_plot_cyclepoints_df(sim_args):

    plot_cyclepoints_df(sim_args['df_samples'], sim_args['sig'], sim_args['fs'], save_fig=True,
                        file_name='test_plot_cyclepoints_df', file_path=TEST_PLOTS_PATH)


@plot_test
def test_plot_cyclepoints_array(sim_args):

    peaks, troughs = find_extrema(sim_args['sig'], sim_args['fs'], (6, 14))
    rises, decays = find_zerox(sim_args['sig'], peaks, troughs)

    plot_cyclepoints_array(sim_args['sig'], sim_args['fs'], peaks=peaks, troughs=troughs,
                           rises=rises, decays=decays, save_fig=True,
                           file_name='test_plot_cyclepoints_array', file_path=TEST_PLOTS_PATH)



"""Tests for features.features."""

import pytest

import numpy as np

from bycycle.features import *

###################################################################################################
###################################################################################################

@pytest.mark.parametrize("burst_method",
    [
        'cycles',
        'amp',
        pytest.param(None, marks=pytest.mark.xfail)
    ]
)
@pytest.mark.parametrize("return_samples", [True, False])
def test_compute_features(sim_args, return_samples, burst_method):

    sig = sim_args['sig']
    fs = sim_args['fs']
    f_range = sim_args['f_range']

    # Set burst detection kwargs
    if burst_method == 'amp':

        threshold_kwargs = {'burst_fraction_threshold': 1, 'min_n_cycles': 3}

    else:

        threshold_kwargs = sim_args['threshold_kwargs']

    # Test returning sample indices in a separate dataframe
    df_features = compute_features(sig, fs, f_range, burst_method=burst_method, \
        threshold_kwargs=threshold_kwargs, return_samples=return_samples)

    if return_samples:

        df_features, df_samples = df_features[0], df_features[1]
        assert len(df_features) == len(df_samples)

    # Assert that np.nan isn't in dataframe columns
    for _, column in df_features.iteritems():

        assert not np.isnan(column[1:-1]).any()

    if return_samples:

        for _, row in df_samples.iterrows():

            assert not np.isnan(row).any()



"""Utility functions for testing bycycle."""

from functools import wraps

import matplotlib.pyplot as plt

###################################################################################################
###################################################################################################

def plot_test(func):
    """Decorator for simple testing of plotting functions.

    Notes
    -----
    This decorator closes all plots prior to the test.
    After running the test function, it checks an axis was created with data.
    It therefore performs a minimal test - asserting the plots exists, with no accuracy checking.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):

        plt.close('all')

        func(*args, **kwargs)

        ax = plt.gca()
        assert ax.has_data()

    return wrapper



"""Tests for cyclepoints.phase."""

import os

import numpy as np

from bycycle.cyclepoints import find_extrema, find_zerox

from bycycle.cyclepoints.phase import *

###################################################################################################
###################################################################################################

def test_extrema_interpolated_phase(sim_stationary):
    """Test waveform phase estimate."""

    # Load signal
    sig = sim_stationary

    fs = 1000
    f_range = (6, 14)

    # Find peaks and troughs
    peaks, troughs = find_extrema(sig, fs, f_range, boundary=1, first_extrema='peak')

    # Find zerocrossings
    rises, decays = find_zerox(sig, peaks, troughs)

    # Compute phase
    pha = extrema_interpolated_phase(sig, peaks, troughs, rises=rises, decays=decays)

    assert len(pha) == len(sig)
    assert np.all(np.isclose(pha[peaks], 0))
    assert np.all(np.isclose(pha[troughs], -np.pi))
    assert np.all(np.isclose(pha[rises], -np.pi/2))
    assert np.all(np.isclose(pha[decays], np.pi/2))



"""Compute shape features for individual cycles."""

import numpy as np
import pandas as pd

from neurodsp.timefrequency import amp_by_time

from bycycle.utils.checks import check_param
from bycycle.utils import rename_extrema_df
from bycycle.features.cyclepoints import compute_cyclepoints

###################################################################################################
###################################################################################################

def compute_shape_features(sig, fs, f_range, center_extrema='peak', find_extrema_kwargs=None,
                           n_cycles=3, return_samples=True):
    """Compute shape features for each cycle.

    Parameters
    ----------
    sig : 1d array
        Time series.
    fs : float
        Sampling rate, in Hz.
    f_range : tuple of (float, float)
        Frequency range for narrowband signal of interest (Hz).
    center_extrema : {'peak', 'trough'}
        The center extrema in the cycle.

        - 'peak' : cycles are defined trough-to-trough
        - 'trough' : cycles are defined peak-to-peak

    find_extrema_kwargs : dict, optional, default: None
        Keyword arguments for function to find peaks an troughs (:func:`~.find_extrema`)
        to change filter parameters or boundary. By default, it sets the filter length to three
        cycles of the low cutoff frequency (``f_range[0]``).
    n_cycles : int, optional, default: 3
        Length of filter, in number of cycles, at the lower cutoff frequency.
    return_samples : bool, optional, default: True
        Returns samples indices of cyclepoints used for determining features if True.

    Returns
    -------
    df_shape_features : pandas.DataFrame
        Dataframe containing cycle shape features. Each row is one cycle. Columns:

        - ``period`` : period of the cycle
        - ``time_decay`` : time between peak and next trough
        - ``time_rise`` : time between peak and previous trough
        - ``time_peak`` : time between rise and decay zero-crosses
        - ``time_trough`` : duration of previous trough estimated by zero-crossings
        - ``volt_decay`` : voltage change between peak and next trough
        - ``volt_rise`` : voltage change between peak and previous trough
        - ``volt_amp`` : average of rise and decay voltage
        - ``volt_peak`` : voltage at the peak
        - ``volt_trough`` : voltage at the last trough
        - ``time_rdsym`` : fraction of cycle in the rise period
        - ``time_ptsym`` : fraction of cycle in the peak period
        - ``band_amp`` : average analytic amplitude of the oscillation

    df_samples : pandas.DataFrame, optional, default: True
        Dataframe containing sample indices of cyclepoints.
        Columns (listed for peak-centered cycles):

        - ``sample_peak`` : sample of 'sig' at which the peak occurs
        - ``sample_zerox_decay`` : sample of the decaying zero-crossing
        - ``sample_zerox_rise`` : sample of the rising zero-crossing
        - ``sample_last_trough`` : sample of the last trough
        - ``sample_next_trough`` : sample of the next trough

    Notes
    -----
    Peak vs trough centering

        - By default, the first extrema analyzed will be a peak, and the final one a trough.
        - In order to switch the preference, the signal is simply inverted and columns are renamed.
        - Columns are slightly different depending on if ``center_extrema`` is set to 'peak' or
          'trough'.

    Examples
    --------
    Compute shape features:

    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> df_shapes, df_samples = compute_shape_features(sig, fs, f_range=(8, 12))
    """

    # Ensure arguments are within valid ranges
    check_param(fs, 'fs', (0, np.inf))
    check_param(n_cycles, 'n_cycles', (0, np.inf))

    # Set defaults if user input is None
    if find_extrema_kwargs is None:
        find_extrema_kwargs = {'filter_kwargs': {'n_cycles': n_cycles}}

    elif 'first_extrema' in find_extrema_kwargs.keys():
        raise ValueError("This function has been designed to assume that the first extrema "
                         "identified will be a peak. This cannot be overwritten at this time.")

    # Negate signal if set to analyze trough-centered cycles
    if center_extrema == 'peak':
        pass
    elif center_extrema == 'trough':
        sig = -sig
    else:
        raise ValueError('Parameter "center_extrema" must be either "peak" or "trough"')

    # For each cycle, identify the sample of each extrema and zero-crossing
    df_samples = compute_cyclepoints(sig, fs, f_range, **find_extrema_kwargs)

    # Compute durations of period, peaks, and troughs
    period, time_peak, time_trough = compute_durations(df_samples)

    # Compute extrema voltage
    volt_peak, volt_trough = compute_extrema_voltage(df_samples, sig)

    # Compute rise-decay and peak-trough features and characteristics
    sym_features = compute_symmetry(df_samples, sig, period=period,
                                    time_peak=time_peak, time_trough=time_trough)

    # Compute average oscillatory amplitude estimate during cycle
    band_amp = compute_band_amp(df_samples, sig, fs, f_range, n_cycles=n_cycles)

    # Organize shape features into a dataframe
    shape_features = {}
    shape_features['period'] = period
    shape_features['time_peak'] = time_peak
    shape_features['time_trough'] = time_trough
    shape_features['volt_peak'] = volt_peak
    shape_features['volt_trough'] = volt_trough
    shape_features['time_decay'] = sym_features['time_decay']
    shape_features['time_rise'] = sym_features['time_rise']
    shape_features['volt_decay'] = sym_features['volt_decay']
    shape_features['volt_rise'] = sym_features['volt_rise']
    shape_features['volt_amp'] = sym_features['volt_amp']
    shape_features['time_rdsym'] = sym_features['time_rdsym']
    shape_features['time_ptsym'] = sym_features['time_ptsym']
    shape_features['band_amp'] = band_amp
    df_shape_features = pd.DataFrame.from_dict(shape_features)

    # Rename the dataframe if trough centered
    df_shape_features, df_samples = rename_extrema_df(center_extrema, df_samples, df_shape_features)

    if return_samples:
        return df_shape_features, df_samples

    return df_shape_features


def compute_durations(df_samples):
    """Compute the time durations of periods, peaks, and troughs.

    Parameters
    ---------
    df_samples : pandas.DataFrame
        Dataframe containing sample indices of cyclepoints.

    Returns
    -------
    period : 1d array
        The period of each cycle.
    time_peak : 1d array
        Time between peak and next trough.
    time_trough : 1d array
        Time between peak and previous trough.

    Examples
    --------
    Compute the durations of periods, peaks, and troughs:

    >>> from bycycle.features import compute_cyclepoints
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> df_samples = compute_cyclepoints(sig, fs, f_range=(8, 12))
    >>> period, time_peak, time_trough = compute_durations(df_samples)
    """

    period = df_samples['sample_next_trough'] - df_samples['sample_last_trough']
    time_peak = df_samples['sample_zerox_decay'] - df_samples['sample_zerox_rise']
    time_trough = df_samples['sample_zerox_rise'] - df_samples['sample_last_zerox_decay']

    return period, time_peak, time_trough


def compute_extrema_voltage(df_samples, sig):
    """Compute the voltage of peaks and troughs.

    Parameters
    ---------
    df_samples : pandas.DataFrame
        Dataframe containing sample indices of cyclepoints.
    sig : 1d array
        Time series.

    Returns
    -------
    volt_peak : 1d array
        Voltage at the peak.
    volt_trough : 1d array
        Voltage at the last trough.

    Examples
    --------
    Compute the voltage at peaks and troughs:

    >>> from bycycle.features import compute_cyclepoints
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> df_samples = compute_cyclepoints(sig, fs, f_range=(8, 12))
    >>> volt_peak, volt_trough = compute_extrema_voltage(df_samples, sig)
    """

    volt_peak = sig[df_samples['sample_peak']]
    volt_trough = sig[df_samples['sample_last_trough']]

    return volt_peak, volt_trough


def compute_symmetry(df_samples, sig, period=None, time_peak=None, time_trough=None):
    """Compute rise-decay and peak-trough characteristics.

    Parameters
    ---------
    df_samples : pandas.DataFrame
        Dataframe containing sample indices of cyclepoints.
    sig : 1d array
        Time series.
    period : 1d array, optional, default: None
        The period of each cycle.
    time_peak : 1d array, optional, default: None
        Time between peak and next trough.
    time_trough : 1d array, optional, default: None
        Time between peak and previous trough

    Returns
    -------
    sym_features : dict
        Contains 1d arrays of symmetry features. Keys include:

        - time_decay : Time between peak and next trough.
        - time_rise : Time between peak and previous trough.
        - volt_decay : Voltage change between peak and next trough.
        - volt_rise : Voltage change between peak and previous trough.
        - volt_amp : Average of rise and decay voltage.
        - time_rdsym : Fraction of cycle in the rise period.
        - time_ptsym : Fraction of cycle in the peak period.

    Examples
    --------
    Compute cycle symmetry characteristics:

    >>> from bycycle.features import compute_cyclepoints
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> df_samples = compute_cyclepoints(sig, fs, f_range=(8, 12))
    >>> sym_features = compute_symmetry(df_samples, sig)
    """

    # Determine rise and decay characteristics
    sym_features = {}

    time_decay = df_samples['sample_next_trough'] - df_samples['sample_peak']
    time_rise = df_samples['sample_peak'] - df_samples['sample_last_trough']
    volt_decay = sig[df_samples['sample_peak']] - sig[df_samples['sample_next_trough']]
    volt_rise = sig[df_samples['sample_peak']] - sig[df_samples['sample_last_trough']]
    volt_amp = (volt_decay + volt_rise) / 2

    # Compute rise-decay symmetry features
    if period is None or time_peak is None or time_trough is None:
        period, time_peak, time_trough = compute_durations(df_samples)

    time_rdsym = time_rise / period

    # Compute peak-trough symmetry features
    time_ptsym = time_peak / (time_peak + time_trough)

    sym_features['time_decay'] = time_decay
    sym_features['time_rise'] = time_rise
    sym_features['volt_decay'] = volt_decay
    sym_features['volt_rise'] = volt_rise
    sym_features['volt_amp'] = volt_amp
    sym_features['time_rdsym'] = time_rdsym
    sym_features['time_ptsym'] = time_ptsym

    return sym_features


def compute_band_amp(df_samples, sig, fs, f_range, n_cycles=3):
    """Compute the average amplitude of each oscillation.

    Parameters
    ----------
    sig : 1d array
        Time series.
    fs : float
        Sampling rate, in Hz.
    f_range : tuple of (float, float)
        Frequency range for narrowband signal of interest (Hz).
    n_cycles : int, optional, default: 3
        Length of filter, in number of cycles, at the lower cutoff frequency.

    Returns
    -------
    band_amp : 1d array
        Average analytic amplitude of the oscillation.

    Examples
    --------
    Compute the mean amplitude for each cycle:

    >>> from bycycle.features import compute_cyclepoints
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> df_samples = compute_cyclepoints(sig, fs, f_range=(8, 12))
    >>> band_amp = compute_band_amp(df_samples, sig, fs, f_range=(8, 12))
    """

    # Ensure arguments are within valid ranges
    check_param(fs, 'fs', (0, np.inf))
    check_param(n_cycles, 'n_cycles', (0, np.inf))

    amp = amp_by_time(sig, fs, f_range, remove_edges=False, n_cycles=n_cycles)

    troughs = np.append(df_samples['sample_last_trough'].values[0],
                        df_samples['sample_next_trough'].values)

    band_amp = [np.mean(amp[troughs[sig_idx]:troughs[sig_idx + 1]]) for
                sig_idx in range(len(df_samples['sample_peak']))]

    return band_amp



"""Find extrema for individual cycles."""

import numpy as np

from neurodsp.filt import filter_signal
from neurodsp.filt.fir import compute_filter_length

from bycycle.utils.checks import check_param
from bycycle.cyclepoints.zerox import find_flank_zerox

###################################################################################################
###################################################################################################

def find_extrema(sig, fs, f_range, boundary=0, first_extrema='peak',
                 filter_kwargs=None, pass_type='bandpass', pad=True):
    """Identify peaks and troughs in a time series.

    Parameters
    ----------
    sig : 1d array
        Time series.
    fs : float
        Sampling rate, in Hz.
    f_range : tuple of (float, float)
        Frequency range, in Hz, to narrowband filter the signal, used to find zero-crossings.
    boundary : int, optional, default: 0
        Number of samples from edge of the signal to ignore.
    first_extrema: {'peak', 'trough', None}, optional, default: 'peak'
        If 'peak', then force the output to begin with a peak and end in a trough.
        If 'trough', then force the output to begin with a trough and end in peak.
        If None, force nothing.
    filter_kwargs : dict, optional, default: None
        Keyword arguments to :func:`~neurodsp.filt.filter.filter_signal`,
        such as 'n_cycles' or 'n_seconds' to control filter length.
    pass_type : str, optional, default: 'bandpass'
        Which kind of filter pass_type is consistent with the frequency definition provided.
    pad : bool, optional, default: True
        Whether to pad ``sig`` with zeroes to prevent missed cyclepoints at the edges.

    Returns
    -------
    peaks : 1d array
        Indices at which oscillatory peaks occur in the input ``sig``.
    troughs : 1d array
        Indices at which oscillatory troughs occur in the input ``sig``.

    Notes
    -----
    This function assures that there are the same number of peaks and troughs
    if the first extrema is forced to be either peak or trough.

    Examples
    --------
    Find the locations of peaks and burst in a signal:

    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> peaks, troughs = find_extrema(sig, fs, f_range=(8, 12))
    """

    # Ensure arguments are within valid range
    check_param(fs, 'fs', (0, np.inf))

    # Set default filtering parameters
    if filter_kwargs is None:
        filter_kwargs = {}

    # Get the original signal and filter lengths
    sig_len = len(sig)
    filt_len = 0

    # Pad beginning of signal with zeros to prevent missing cyclepoints
    if pad:

        filt_len = compute_filter_length(fs, pass_type, f_range[0], f_range[1],
                                         n_seconds=filter_kwargs.get('n_seconds', None),
                                         n_cycles=filter_kwargs.get('n_cycles', 3))

        # Pad the signal
        sig = np.pad(sig, int(np.ceil(filt_len/2)), mode='constant')

    # Narrowband filter signal
    sig_filt = filter_signal(sig, fs, pass_type, f_range, remove_edges=False, **filter_kwargs)

    # Find rising and decaying zero-crossings (narrowband)
    rise_xs = find_flank_zerox(sig_filt, 'rise')
    decay_xs = find_flank_zerox(sig_filt, 'decay')

    # Compute number of peaks and troughs
    if rise_xs[-1] > decay_xs[-1]:
        n_peaks = len(rise_xs) - 1
        n_troughs = len(decay_xs)
    else:
        n_peaks = len(rise_xs)
        n_troughs = len(decay_xs) - 1

    # Calculate peak samples
    peaks = np.zeros(n_peaks, dtype=int)
    for p_idx in range(n_peaks):

        # Calculate the sample range between the most recent zero rise and the next zero decay
        last_rise = rise_xs[p_idx]
        next_decay = decay_xs[decay_xs > last_rise][0]
        # Identify time of peak
        peaks[p_idx] = np.argmax(sig[last_rise:next_decay]) + last_rise

    # Calculate trough samples
    troughs = np.zeros(n_troughs, dtype=int)
    for t_idx in range(n_troughs):

        # Calculate the sample range between the most recent zero decay and the next zero rise
        last_decay = decay_xs[t_idx]
        next_rise = rise_xs[rise_xs > last_decay][0]
        # Identify time of trough
        troughs[t_idx] = np.argmin(sig[last_decay:next_rise]) + last_decay

    # Remove padding
    peaks = peaks - int(np.ceil(filt_len/2))
    troughs = troughs - int(np.ceil(filt_len/2))

    # Remove peaks and troughs within the boundary limit
    peaks = peaks[np.logical_and(peaks > boundary, peaks < sig_len - boundary)]
    troughs = troughs[np.logical_and(troughs > boundary, troughs < sig_len - boundary)]

    # Force the first extrema to be as desired & assure equal # of peaks and troughs
    if first_extrema == 'peak':
        troughs = troughs[1:] if peaks[0] > troughs[0] else troughs
        peaks = peaks[:-1] if peaks[-1] > troughs[-1] else peaks
    elif first_extrema == 'trough':
        peaks = peaks[1:] if troughs[0] > peaks[0] else peaks
        troughs = troughs[:-1] if troughs[-1] > peaks[-1] else troughs
    elif first_extrema is None:
        pass
    else:
        raise ValueError('Parameter "first_extrema" is invalid')

    return peaks, troughs



"""ByCycle setup script."""

import os
from setuptools import setup, find_packages

# Get the current version number from inside the module
with open(os.path.join('bycycle', 'version.py')) as version_file:
    exec(version_file.read())

# Load the long description from the README
with open('README.rst') as readme_file:
    long_description = readme_file.read()

# Load the required dependencies from the requirements file
with open("requirements.txt") as requirements_file:
    install_requires = [req for req in requirements_file.read().splitlines()]

setup(
    name = 'bycycle',
    version = __version__,
    description = 'Cycle-by-cycle analyses of neural oscillations.',
    long_description = long_description,
    python_requires = '>=3.5',
    author = 'The Voytek Lab',
    author_email = 'voyteklab@gmail.com',
    url = 'https://github.com/bycycle-tools/bycycle',
    packages = find_packages(),
    license = 'Apache License, 2.0',
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    platforms = 'any',
    project_urls = {
        'Documentation' : 'https://bycycle-tools.github.io/bycycle/',
        'Bug Reports' : 'https://github.com/bycycle-tools/bycycle/issues',
        'Source' : 'https://github.com/bycycle-tools/bycycle'
    },
    download_url = 'https://github.com/bycycle-tools/bycycle/releases',
    keywords = ['neuroscience', 'neural oscillations', 'waveform', 'shape', 'electrophysiology'],
    install_requires = install_requires,
    tests_require = ['pytest']
)



"""Tests for features.cyclepoints."""

from bycycle.cyclepoints import find_extrema, find_zerox

from bycycle.features.cyclepoints import *

###################################################################################################
###################################################################################################

def test_compute_cyclepoints(sim_args):

    sig = sim_args['sig']
    fs = sim_args['fs']
    f_range = sim_args['f_range']

    peaks, troughs = find_extrema(sig, fs, f_range)
    rises, decays = find_zerox(sig, peaks, troughs)

    df_samples = compute_cyclepoints(sig, fs, f_range)

    assert (df_samples['sample_peak'] == peaks[1:]).all()
    assert (df_samples['sample_zerox_decay'] == decays[1:]).all()
    assert (df_samples['sample_zerox_rise'] == rises).all()
    assert (df_samples['sample_last_trough'] == troughs[:-1]).all()
    assert (df_samples['sample_next_trough'] == troughs[1:]).all()



"""Tests for features.shape."""

import pytest

import numpy as np

from bycycle.cyclepoints import find_extrema, find_zerox

from bycycle.features.shape import *

###################################################################################################
###################################################################################################

@pytest.mark.parametrize("find_extrema_kwargs",
    [
        None,
        pytest.param({'first_extrema': 'peak'}, marks=pytest.mark.xfail)
    ]
)
@pytest.mark.parametrize("center_extrema",
    [
        'peak',
        'trough',
        pytest.param(None, marks=pytest.mark.xfail)
    ]
)
@pytest.mark.parametrize("return_samples", [True, False])
def test_compute_shape_features(sim_args, find_extrema_kwargs, center_extrema, return_samples):

    sig = sim_args['sig']
    fs = sim_args['fs']
    f_range = sim_args['f_range']

    outs = compute_shape_features(sig, fs, f_range,
                                  center_extrema=center_extrema,
                                  find_extrema_kwargs=find_extrema_kwargs,
                                  return_samples=return_samples)

    if return_samples:
        df_shapes, df_samples = outs
        assert len(df_shapes) == len(df_samples)

    else:
        df_shapes = outs

    # Assert that np.nan isn't dataframe(s), with the exception of the first and last row
    for idx, row in df_shapes.iterrows():

        assert not np.isnan(row[1:-1]).any()

        if return_samples:

            assert not np.isnan(df_samples.iloc[idx]).any()

    # Check inverted signal gives appropriately opposite data
    extrema_opp = 'trough' if center_extrema == 'peak' else 'peak'

    df_opp = compute_shape_features(-sig, fs, f_range,
                                    center_extrema=extrema_opp,
                                    find_extrema_kwargs=find_extrema_kwargs,
                                    return_samples=False)

    cols_peak = ['time_peak', 'time_rise', 'volt_rise',
                 'volt_amp', 'period', 'time_rdsym', 'time_ptsym']
    cols_trough = ['time_trough', 'time_decay', 'volt_decay',
                   'volt_amp', 'period', 'time_rdsym', 'time_ptsym']

    df_opp['time_rdsym'] = 1 - df_opp['time_rdsym']
    df_opp['time_ptsym'] = 1 - df_opp['time_ptsym']

    for idx, col in enumerate(cols_peak):

        if center_extrema == 'peak':
            np.testing.assert_allclose(df_shapes.loc[:, col], df_opp.loc[:, cols_trough[idx]])
        else:
            np.testing.assert_allclose(df_opp.loc[:, col], df_shapes.loc[:, cols_trough[idx]])


def test_compute_durations(sim_args):

    df_samples = sim_args['df_samples']
    period, time_peak, time_trough = compute_durations(df_samples)

    assert ((time_trough + time_peak) == period).all()


def test_compute_extrema_voltage(sim_args):

    df_samples = sim_args['df_samples']
    sig = sim_args['sig']

    volt_peaks, volt_troughs = compute_extrema_voltage(df_samples, sig)

    for idx, volt_peak in enumerate(volt_peaks):
        assert volt_peak > volt_troughs[idx]


@pytest.mark.parametrize("recompute", [True, False])
def test_compute_symmetry(sim_args, recompute):

    df_samples = sim_args['df_samples']
    sig = sim_args['sig']

    if recompute:
        sym_features = compute_symmetry(df_samples, sig)
    else:
        period, time_peak, time_trough = compute_durations(df_samples)
        sym_features = compute_symmetry(df_samples, sig, period=period,
                                        time_peak=time_peak, time_trough=time_trough)

    # This is the case for only simulated sine waves
    assert (sym_features['time_decay'] == sym_features['time_rise']).all()
    assert (sym_features['volt_decay'] == sym_features['volt_rise']).all()

    assert (sym_features['volt_amp'] == \
        (sym_features['volt_rise'] + sym_features['volt_decay'])/2).all()

    np.testing.assert_almost_equal(sym_features['time_rdsym'].values,
                                   sym_features['time_ptsym'].values, decimal=1)


def test_compute_band_amp(sim_args):

    df_samples = sim_args['df_samples']
    sig = sim_args['sig']
    fs = sim_args['fs']
    f_range = sim_args['f_range']

    band_amp = compute_band_amp(df_samples, sig, fs, f_range)

    for amp in band_amp[:-1]:
        np.testing.assert_allclose(amp, band_amp[0], rtol=1e-1)



"""Detect bursts: cycle consistency approach."""

import pandas as pd

from bycycle.utils.checks import check_param
from bycycle.burst.utils import check_min_burst_cycles

###################################################################################################
###################################################################################################

pd.options.mode.chained_assignment = None

def detect_bursts_cycles(df_features, amp_fraction_threshold=0., amp_consistency_threshold=.5,
                         period_consistency_threshold=.5, monotonicity_threshold=.8,
                         min_n_cycles=3):
    """Detects bursts based on consistency between cycles.

    Parameters
    ----------
    df_features : pandas.DataFrame
        Waveform features for individual cycles from :func:`~.compute_burst_features`.
    amp_fraction_threshold : float, optional, default: 0.
        The minimum normalized amplitude a cycle must have in order to be considered in an
        oscillation. Must be between 0 and 1.

        - 0 = the minimum amplitude across all cycles
        - .5 = the median amplitude across all cycles
        - 1 = the maximum amplitude across all cycles

    amp_consistency_threshold : float, optional, default: 0.5
        The minimum normalized difference in rise and decay magnitude to be considered as in an
        oscillatory mode. Must be between 0 and 1.

        - 1 = the same amplitude for the rise and decay
        - .5 = the rise (or decay) is half the amplitude of the decay (rise)

    period_consistency_threshold : float, optional, default: 0.5
        The minimum normalized difference in period between two adjacent cycles to be considered
        as in an oscillatory mode. Must be between 0 and 1.

        - 1 = the same period for both cycles
        - .5 = one cycle is half the duration of another cycle

    monotonicity_threshold : float, optional, default: 0.8
        The minimum fraction of time segments between samples that must be going in the same
        direction. Must be between 0 and 1.

        - 1 = rise and decay are perfectly monotonic
        - .5 = both rise and decay are rising half of the time and decay half the time
        - 0 = rise period is all decaying and decay period is all rising

    min_n_cycles : int, optional, default: 3
        The minimum number of cycles of consecutive cycles required to be considered a burst.

    Returns
    -------
    df_features : pandas.DataFrame
        Same df as input, with an additional column (`is_burst`) to indicate if the cycle is part
        of an oscillatory burst, with additional columns indicating the burst detection parameters.

    Notes
    -----
    * The first and last period cannot be considered oscillating if the consistency measures are
      used.

    Examples
    --------
    Apply thresholding for consistency burst detection:

    >>> from bycycle.features import compute_burst_features, compute_shape_features
    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> df_shapes, df_samples = compute_shape_features(sig, fs, f_range=(8, 12))
    >>> df_burst = compute_burst_features(df_shapes, df_samples, sig)
    >>> df_burst = detect_bursts_cycles(df_burst)
    """

    # Ensure arguments are within valid ranges
    check_param(amp_fraction_threshold, 'amp_fraction_threshold', (0, 1))
    check_param(amp_consistency_threshold, 'amp_consistency_threshold', (0, 1))
    check_param(period_consistency_threshold, 'period_consistency_threshold', (0, 1))
    check_param(monotonicity_threshold, 'monotonicity_threshold', (0, 1))

    # Compute if each period is part of an oscillation
    amp_fraction = df_features['amp_fraction'] > amp_fraction_threshold
    amp_consistency = df_features['amp_consistency'] > amp_consistency_threshold
    period_consistency = df_features['period_consistency'] > period_consistency_threshold
    monotonicity = df_features['monotonicity'] > monotonicity_threshold

    # Set the burst status for each cycle as the answer across all criteria
    is_burst = amp_fraction & amp_consistency & period_consistency & monotonicity

    # Set the first and last cycles to not be part of a burst
    is_burst[0] = False
    is_burst[-1] = False

    df_features['is_burst'] = check_min_burst_cycles(is_burst, min_n_cycles=min_n_cycles)

    return df_features



"""Tests for cyclepoints.extrema."""

import os

import numpy as np
from scipy.signal import argrelextrema

import pytest

from bycycle.cyclepoints.extrema import *
from bycycle.tests.settings import FS, F_RANGE

###################################################################################################
###################################################################################################

@pytest.mark.parametrize("first_extrema",
    [
        'peak',
        'trough',
        None,
        pytest.param('fail', marks=pytest.mark.xfail(raises=ValueError))
    ]
)
def test_find_extrema(sim_args, first_extrema):
    """Test ability to find peaks and troughs."""

    # Load signal
    sig = sim_args['sig']

    # find local maxima and minima using scipy
    maxima = argrelextrema(sig, np.greater)
    minima = argrelextrema(sig, np.less)

    # Find peaks and troughs using bycycle and make sure match scipy
    peaks, troughs = find_extrema(sig, FS, F_RANGE, first_extrema=first_extrema)

    if first_extrema == 'trough':
        assert troughs[0] < peaks[0]
        assert len(peaks) == len(troughs)

        # The first extrema is actually a peak; slice maxima to account for this when
        #   trough is forced as the first extrema.
        np.testing.assert_equal(peaks, maxima[0][1:])

        # When first_extrema is trough, the last extrema must be a peak. This leads to a missing
        #   last trough.
        np.testing.assert_equal(troughs, minima[0][:-1])

    elif first_extrema == 'peak':
        assert peaks[0] < troughs[0]
        assert len(peaks) == len(troughs)
        np.testing.assert_equal(peaks, maxima[0])
        np.testing.assert_equal(troughs, minima[0])







"""Find zero-crossings for individual cycles."""

from operator import gt, lt

import numpy as np

###################################################################################################
###################################################################################################

def find_zerox(sig, peaks, troughs):
    """Find zero-crossings within each cycle, from identified peaks and troughs.

    Parameters
    ----------
    sig : 1d array
        Time series.
    peaks : 1d array
        Samples of oscillatory peaks.
    troughs : 1d array
        Samples of oscillatory troughs.

    Returns
    -------
    rises : 1d array
        Samples at which oscillatory rising zero-crossings occur.
    decays : 1d array
        Samples at which oscillatory decaying zero-crossings occur.

    Notes
    -----
    - Zero-crossings are defined as when the voltage crosses midway between one extrema and
      the next - for example, a 'rise' is halfway from the trough to the peak.
    - If this halfway voltage is crossed at multiple times, the temporal median is taken
      as the zero-crossing.
    - Sometimes, due to noise in estimating peaks and troughs when the oscillation
      is absent, the estimated peak might be lower than an adjacent trough. If this
      occurs, the rise and decay zero-crossings will be set to be halfway between
      the peak and trough.
    - Burst detection should be used to restrict phase estimation to periods with oscillations
      present, in order to ignore periods of the signal in which estimation is poor.

    Examples
    --------
    Find the rise and decay zero-crossings locations of a simulated signal:

    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> from bycycle.cyclepoints import find_extrema
    >>> fs = 500
    >>> sig = sim_bursty_oscillation(10, fs, freq=10)
    >>> peaks, troughs = find_extrema(sig, fs, f_range=(8, 12))
    >>> rises, decays = find_zerox(sig, peaks, troughs)
    """

    # Calculate the number of rises and decays
    n_rises = len(peaks)
    n_decays = len(troughs)
    idx_bias = 0

    # Offset values, depending on order of peaks & troughs
    if peaks[0] < troughs[0]:
        n_rises -= 1
    else:
        n_decays -= 1
        idx_bias += 1

    rises = _find_flank_midpoints(sig, 'rise', n_rises, troughs, peaks, idx_bias)
    decays = _find_flank_midpoints(sig, 'decay', n_decays, peaks, troughs, idx_bias)

    return rises, decays


def find_flank_zerox(sig, flank):
    """Find zero-crossings on rising or decay flanks of a filtered signal.

    Parameters
    ----------
    sig : 1d array
        Time series to detect zero-crossings in.
    flank : {'rise', 'decay'}
        Which flank, rise or decay, to use to get zero crossings.

    Returns
    -------
    zero_xs : 1d array
        Samples of the zero crossings.

    Examples
    --------
    Find rising flanks in a filtered signal:

    >>> from neurodsp.sim import sim_bursty_oscillation
    >>> from neurodsp.filt import filter_signal
    >>> sig = sim_bursty_oscillation(10, 500, freq=10)
    >>> sig_filt = filter_signal(sig, 500, 'lowpass', 30)
    >>> rises_flank = find_flank_zerox(sig_filt, 'rise')
    """

    assert flank in ['rise', 'decay']
    pos = sig <= 0 if flank == 'rise' else sig > 0

    zero_xs = (pos[:-1] & ~pos[1:]).nonzero()[0]

    # If no zero-crossing's found (peak and trough are same voltage), output dummy value
    zero_xs = [int(len(sig) / 2)] if len(zero_xs) == 0 else zero_xs

    return zero_xs


def _find_flank_midpoints(sig, flank, n_flanks, extrema_start, extrema_end, idx_bias):
    """Helper function for find_zerox."""

    assert flank in ['rise', 'decay']
    idx_bias = -idx_bias + 1 if flank == 'rise' else idx_bias
    comp = gt if flank == 'rise' else lt

    flanks = np.zeros(n_flanks, dtype=int)
    for idx in range(n_flanks):

        sig_temp = np.copy(sig[extrema_start[idx]:extrema_end[idx + idx_bias] + 1])
        sig_temp -= (sig_temp[0] + sig_temp[-1]) / 2.

        # If data is all zeros, just set the zero-crossing to be halfway between
        if np.sum(np.abs(sig_temp)) == 0:
            flanks[idx] = extrema_start[idx] + int(len(sig_temp) / 2.)

        # If flank is actually an extrema, just set the zero-crossing to be halfway between
        elif comp(sig_temp[0], sig_temp[-1]):
            flanks[idx] = extrema_start[idx] + int(len(sig_temp) / 2.)

        else:
            flanks[idx] = extrema_start[idx] + int(np.median(find_flank_zerox(sig_temp, flank)))

    return flanks
